import collections
from typing import (
    Dict,
    Set,
    Union,
    cast,
)
import uuid

from eth_utils import (
    ValidationError,
)
from eth_utils.toolz import (
    first,
    merge,
    nth,
)

from eth.abc import (
    DatabaseAPI,
)
from eth.db.backends.base import (
    BaseDB,
)
from eth.db.diff import (
    DBDiff,
    DBDiffTracker,
)


class DeletedEntry:
    pass


# Track two different kinds of deletion:

# 1. key in wrapped
# 2. key modified in journal
# 3. key deleted
DELETED_ENTRY = DeletedEntry()

# 1. key not in wrapped
# 2. key created in journal
# 3. key deleted
ERASE_CREATED_ENTRY = DeletedEntry()


class Journal(BaseDB):
    """
    A Journal is an ordered list of changesets.  A changeset is a dictionary
    of database keys and values.  The values are tracked changes that were
    written after the changeset was created

    Changesets are referenced by a random uuid4.
    """

    def __init__(self) -> None:
        # contains a mapping from all of the `uuid4` changeset_ids
        # to a dictionary of key:value pairs with the recorded changes
        # that belong to the changeset
        self.journal_data: collections.OrderedDict[
            uuid.UUID, Dict[bytes, Union[bytes, DeletedEntry]]
        ] = collections.OrderedDict()
        self._clears_at: Set[uuid.UUID] = set()

    @property
    def root_changeset_id(self) -> uuid.UUID:
        """
        Returns the id of the root changeset
        """
        return first(self.journal_data.keys())

    @property
    def is_flattened(self) -> bool:
        """
        :return: whether there are any explicitly committed checkpoints
        """
        return len(self.journal_data) < 2

    @property
    def latest_id(self) -> uuid.UUID:
        """
        Returns the id of the latest changeset
        """
        # last() was iterating through all values, so first(reversed()) gives
        # a 12.5x speedup
        return first(reversed(self.journal_data.keys()))

    @property
    def latest(self) -> Dict[bytes, Union[bytes, DeletedEntry]]:
        """
        Returns the dictionary of db keys and values for the latest changeset.
        """
        return self.journal_data[self.latest_id]

    @latest.setter
    def latest(self, value: Dict[bytes, Union[bytes, DeletedEntry]]) -> None:
        """
        Setter for updating the *latest* changeset.
        """
        self.journal_data[self.latest_id] = value

    def is_empty(self) -> bool:
        return len(self.journal_data) == 0

    def has_changeset(self, changeset_id: uuid.UUID) -> bool:
        return changeset_id in self.journal_data

    def record_changeset(self, custom_changeset_id: uuid.UUID = None) -> uuid.UUID:
        """
        Creates a new changeset. Changesets are referenced by a random uuid4
        to prevent collisions between multiple changesets.
        """
        if custom_changeset_id is not None:
            if custom_changeset_id in self.journal_data:
                raise ValidationError(
                    "Tried to record with an existing "
                    f"changeset id: {custom_changeset_id!r}"
                )
            else:
                changeset_id = custom_changeset_id
        else:
            changeset_id = uuid.uuid4()

        self.journal_data[changeset_id] = {}
        return changeset_id

    def pop_changeset(
        self, changeset_id: uuid.UUID
    ) -> Dict[bytes, Union[bytes, DeletedEntry]]:
        """
        Returns all changes from the given changeset.  This includes all of
        the changes from any subsequent changeset, giving precedence to
        later changesets.
        """
        if changeset_id not in self.journal_data:
            raise KeyError(changeset_id, "Unknown changeset in JournalDB")

        all_ids = tuple(self.journal_data.keys())
        changeset_idx = all_ids.index(changeset_id)
        changesets_to_pop = all_ids[changeset_idx:]
        popped_clears = tuple(
            idx for idx in changesets_to_pop if idx in self._clears_at
        )
        if popped_clears:
            last_clear_idx = changesets_to_pop.index(popped_clears[-1])
            changesets_to_drop = changesets_to_pop[:last_clear_idx]
            changesets_to_merge = changesets_to_pop[last_clear_idx:]
        else:
            changesets_to_drop = ()
            changesets_to_merge = changesets_to_pop

        # we pull all of the changesets *after* the changeset we are
        # reverting to and collapse them to a single set of keys (giving
        # precedence to later changesets)
        changeset_data = merge(
            *(self.journal_data.pop(c_id) for c_id in changesets_to_merge)
        )

        # drop the changes on the floor if they came before a clear that is
        # being committed
        for changeset_id in changesets_to_drop:
            self.journal_data.pop(changeset_id)

        self._clears_at.difference_update(popped_clears)
        return changeset_data

    def clear(self) -> None:
        """
        Treat as if the *underlying* database will also be cleared by some other
        mechanism. We build a special empty changeset just for marking that all
        previous data should be ignored.
        """
        # these internal records are used as a way to tell the difference between
        # changes that came before and after the clear
        self.record_changeset()
        self._clears_at.add(self.latest_id)
        self.record_changeset()

    def has_clear(self, check_changeset_id: uuid.UUID) -> bool:
        for changeset_id in reversed(self.journal_data.keys()):
            if changeset_id in self._clears_at:
                return True
            elif check_changeset_id == changeset_id:
                return False
        raise ValidationError(
            "Changeset ID %s is not in the journal" % check_changeset_id
        )

    def commit_changeset(
        self, changeset_id: uuid.UUID
    ) -> Dict[bytes, Union[bytes, DeletedEntry]]:
        """
        Collapses all changes for the given changeset into the previous
        changesets if it exists.
        """
        does_clear = self.has_clear(changeset_id)
        changeset_data = self.pop_changeset(changeset_id)
        if not self.is_empty():
            # we only have to assign changeset data into the latest changeset if
            # there is one.
            if does_clear:
                # if there was a clear and more changesets underneath then clear the
                # latest changeset, and replace with a new clear changeset
                self.latest = {}
                self._clears_at.add(self.latest_id)
                self.record_changeset()
                self.latest = changeset_data
            else:
                # otherwise, merge in all the current data
                self.latest = merge(
                    self.latest,
                    changeset_data,
                )
        return changeset_data

    def flatten(self) -> None:
        if self.is_flattened:
            return

        changeset_id_after_root = nth(1, self.journal_data.keys())
        self.commit_changeset(changeset_id_after_root)

    #
    # Database API
    #
    def __getitem__(self, key: bytes) -> Union[bytes, DeletedEntry]:  # type: ignore # Breaks LSP  # noqa: E501
        """
        For key lookups we need to iterate through the changesets in reverse
        order, returning from the first one in which the key is present.
        """
        # Ignored from mypy because of https://github.com/python/typeshed/issues/2078
        for changeset_id, changeset_data in reversed(self.journal_data.items()):
            if changeset_id in self._clears_at:
                return ERASE_CREATED_ENTRY
            elif key in changeset_data:
                return changeset_data[key]
            else:
                continue

        return None

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.latest[key] = value

    def _exists(self, key: bytes) -> bool:
        val = self.get(key)
        return val is not None and val not in (ERASE_CREATED_ENTRY, DELETED_ENTRY)

    def __delitem__(self, key: bytes) -> None:
        raise NotImplementedError(
            "You must delete with one of delete_local or delete_wrapped"
        )

    def delete_wrapped(self, key: bytes) -> None:
        self.latest[key] = DELETED_ENTRY

    def delete_local(self, key: bytes) -> None:
        self.latest[key] = ERASE_CREATED_ENTRY

    def diff(self) -> DBDiff:
        tracker = DBDiffTracker()
        visited_keys: Set[bytes] = set()

        # Iterate in reverse, so you can skip over any keys from old checkpoints.
        # This is required so that when a key is created and then deleted in the
        #   journal, we don't add the delete to the diff.
        #   (We simply omit the change altogether)
        for changeset_id, changeset in reversed(self.journal_data.items()):
            if changeset_id in self._clears_at:
                break

            for key, value in changeset.items():
                if key in visited_keys:
                    # this old change has already been tracked
                    continue
                elif value is DELETED_ENTRY:
                    del tracker[key]
                elif value is ERASE_CREATED_ENTRY:
                    pass
                else:
                    tracker[key] = cast(bytes, value)

                visited_keys.add(key)

        return tracker.diff()


class JournalDB(BaseDB):
    """
    A wrapper around the basic DB objects that keeps a journal of all changes.
    Each time a recording is started, the underlying journal creates a new
    changeset and assigns an id to it. The journal then keeps track of all changes
    that go into this changeset.

    Discarding a changeset simply throws it away inculding all subsequent changesets
    that may have followed. Commiting a changeset merges the given changeset and all
    subsequent changesets into the previous changeset giving precidence to later
    changesets in case of conflicting keys.

    Nothing is written to the underlying db until `persist()` is called.

    The added memory footprint for a JournalDB is one key/value stored per
    database key which is changed.  Subsequent changes to the same key within
    the same changeset will not increase the journal size since we only need
    to track latest value for any given key within any given changeset.
    """

    wrapped_db = None
    journal: Journal = None

    def __init__(self, wrapped_db: DatabaseAPI) -> None:
        self.wrapped_db = wrapped_db
        self.reset()

    def __getitem__(self, key: bytes) -> bytes:
        val = self.journal[key]
        if val is DELETED_ENTRY:
            raise KeyError(
                key,
                "item is deleted in JournalDB, and will be deleted from the wrapped DB",
            )
        elif val is ERASE_CREATED_ENTRY:
            raise KeyError(
                key,
                "item is deleted in JournalDB, and is presumed gone "
                "from the wrapped DB",
            )
        elif val is None:
            return self.wrapped_db[key]
        else:
            # mypy doesn't allow custom type guards yet so we need to cast here
            # even though we know it can only be `bytes` at this point.
            return cast(bytes, val)

    def __setitem__(self, key: bytes, value: bytes) -> None:
        """
        - replacing an existing value
        - setting a value that does not exist
        """
        self.journal[key] = value

    def _exists(self, key: bytes) -> bool:
        val = self.journal[key]
        if val in (ERASE_CREATED_ENTRY, DELETED_ENTRY):
            return False
        elif val is None:
            return key in self.wrapped_db
        else:
            return True

    def clear(self) -> None:
        """
        Remove all keys. Immediately after a clear, *all* getitem requests will
        return a KeyError. That includes the changes pending persist and any data
        in the underlying database.

        (This action is journaled, like all other actions)

        clear will *not* persist the emptying of all keys in the underlying DB.
        It only prevents any updates (or deletes!) before it from being persisted.

        Any caller that wants to use clear must also make sure that the underlying
        database reflects their desired end state (maybe emptied, maybe not).
        """
        self.journal.clear()

    def has_clear(self) -> bool:
        return self.journal.has_clear(self.journal.root_changeset_id)

    def __delitem__(self, key: bytes) -> None:
        if key in self.wrapped_db:
            self.journal.delete_wrapped(key)
        else:
            if key in self.journal:
                self.journal.delete_local(key)
            else:
                raise KeyError(
                    key, "key could not be deleted in JournalDB, because it was missing"
                )

    #
    # Snapshot API
    #
    def _validate_changeset(self, changeset_id: uuid.UUID) -> None:
        """
        Checks to be sure the changeset is known by the journal
        """
        if not self.journal.has_changeset(changeset_id):
            raise ValidationError(
                f"Changeset not found in journal: {str(changeset_id)}"
            )

    def has_changeset(self, changeset_id: uuid.UUID) -> bool:
        return self.journal.has_changeset(changeset_id)

    def record(self, custom_changeset_id: uuid.UUID = None) -> uuid.UUID:
        """
        Starts a new recording and returns an id for the associated changeset
        """
        return self.journal.record_changeset(custom_changeset_id)

    def discard(self, changeset_id: uuid.UUID) -> None:
        """
        Throws away all journaled data starting at the given changeset
        """
        self._validate_changeset(changeset_id)
        self.journal.pop_changeset(changeset_id)

    def commit(self, changeset_id: uuid.UUID) -> None:
        """
        Commits a given changeset. This merges the given changeset and all
        subsequent changesets into the previous changeset giving precidence
        to later changesets in case of any conflicting keys.
        """
        self._validate_changeset(changeset_id)
        if changeset_id == self.journal.root_changeset_id:
            raise ValidationError(
                "Tried to commit the root changeset. Callers should not keep references"
                " to the root changeset. Maybe you meant to use persist()?"
            )
        self.journal.commit_changeset(changeset_id)

    def _reapply_changeset_to_journal(
        self,
        changeset_id: uuid.UUID,
        journal_data: Dict[bytes, Union[bytes, DeletedEntry]],
    ) -> None:
        self.record(changeset_id)
        for key, value in journal_data.items():
            if value is DELETED_ENTRY:
                self.journal.delete_wrapped(key)
            elif value is ERASE_CREATED_ENTRY:
                self.journal.delete_local(key)
            else:
                self.journal[key] = cast(bytes, value)

    def persist(self) -> None:
        """
        Persist all changes in underlying db. After all changes have been written the
        JournalDB starts a new recording.
        """
        root_changeset = self.journal.root_changeset_id
        journal_data = self.journal.commit_changeset(root_changeset)

        # Ensure the journal automatically restarts recording after
        # it has been persisted to the underlying db
        self.reset()

        for key, value in journal_data.items():
            try:
                if value is DELETED_ENTRY:
                    del self.wrapped_db[key]
                elif value is ERASE_CREATED_ENTRY:
                    pass
                else:
                    self.wrapped_db[key] = cast(bytes, value)
            except Exception:
                self._reapply_changeset_to_journal(root_changeset, journal_data)
                raise

    def flatten(self) -> None:
        """
        Commit everything possible without persisting
        """
        self.journal.flatten()

    def reset(self) -> None:
        """
        Reset the entire journal.
        """
        self.journal = Journal()
        self.record()

    def diff(self) -> DBDiff:
        """
        Generate a DBDiff of all pending changes.
        These are the changes that would occur if :meth:`persist()` were called.
        """
        return self.journal.diff()



from typing import (
    Type,
)

from eth.abc import (
    TransactionExecutorAPI,
)

from ..london import (
    LondonState,
)
from ..london.state import (
    LondonTransactionExecutor,
)
from .computation import (
    ArrowGlacierComputation,
)


class ArrowGlacierTransactionExecutor(LondonTransactionExecutor):
    pass


class ArrowGlacierState(LondonState):
    computation_class = ArrowGlacierComputation
    transaction_executor_class: Type[
        TransactionExecutorAPI
    ] = ArrowGlacierTransactionExecutor



from contextlib import (
    contextmanager,
)
import logging
from typing import (
    FrozenSet,
    Iterator,
    Set,
)

from eth.abc import (
    AtomicDatabaseAPI,
    AtomicWriteBatchAPI,
    DatabaseAPI,
)
from eth.db.atomic import (
    BaseAtomicDB,
)
from eth.db.backends.base import (
    BaseDB,
)


class KeyAccessLoggerDB(BaseDB):
    """
    Wraps around a database and tracks all the keys that were read since initialization.
    """

    logger = logging.getLogger("eth.db.KeyAccessLoggerDB")

    def __init__(self, wrapped_db: DatabaseAPI, log_missing_keys: bool = True) -> None:
        """
        :param log_missing_keys: True if a key is added to :attr:`keys_read` even if the
            key/value does not exist in the database.
        """
        self.wrapped_db = wrapped_db
        self._keys_read: Set[bytes] = set()
        self._log_missing_keys = log_missing_keys

    @property
    def keys_read(self) -> FrozenSet[bytes]:
        # Make a defensive copy so callers can't modify the list externally
        return frozenset(self._keys_read)

    def __getitem__(self, key: bytes) -> bytes:
        try:
            result = self.wrapped_db.__getitem__(key)
        except KeyError:
            if self._log_missing_keys:
                self._keys_read.add(key)
            raise
        else:
            self._keys_read.add(key)
            return result

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.wrapped_db[key] = value

    def __delitem__(self, key: bytes) -> None:
        del self.wrapped_db[key]

    def _exists(self, key: bytes) -> bool:
        does_exist = key in self.wrapped_db
        if does_exist or self._log_missing_keys:
            self._keys_read.add(key)
        return does_exist


class KeyAccessLoggerAtomicDB(BaseAtomicDB):
    """
    Wraps around an atomic database and tracks all the keys
    that were read since initialization.
    """

    logger = logging.getLogger("eth.db.KeyAccessLoggerAtomicDB")

    def __init__(
        self, wrapped_db: AtomicDatabaseAPI, log_missing_keys: bool = True
    ) -> None:
        """
        :param log_missing_keys: True if a key is added to :attr:`keys_read` even if the
            key/value does not exist in the database.
        """
        self.wrapped_db = wrapped_db
        self._keys_read: Set[bytes] = set()
        self._log_missing_keys = log_missing_keys

    @property
    def keys_read(self) -> FrozenSet[bytes]:
        # Make a defensive copy so callers can't modify the list externally
        return frozenset(self._keys_read)

    def __getitem__(self, key: bytes) -> bytes:
        try:
            result = self.wrapped_db.__getitem__(key)
        except KeyError:
            if self._log_missing_keys:
                self._keys_read.add(key)
            raise
        else:
            self._keys_read.add(key)
            return result

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.wrapped_db[key] = value

    def __delitem__(self, key: bytes) -> None:
        del self.wrapped_db[key]

    def _exists(self, key: bytes) -> bool:
        does_exist = key in self.wrapped_db
        if does_exist or self._log_missing_keys:
            self._keys_read.add(key)
        return does_exist

    @contextmanager
    def atomic_batch(self) -> Iterator[AtomicWriteBatchAPI]:
        with self.wrapped_db.atomic_batch() as readable_batch:
            yield readable_batch



from typing import (
    Type,
)

from eth.abc import (
    ComputationAPI,
    SignedTransactionAPI,
)
from eth.vm.forks.frontier.state import (
    FrontierState,
    FrontierTransactionExecutor,
)

from .computation import (
    HomesteadComputation,
)
from .validation import (
    validate_homestead_transaction,
)


class HomesteadState(FrontierState):
    computation_class: Type[ComputationAPI] = HomesteadComputation

    def validate_transaction(self, transaction: SignedTransactionAPI) -> None:
        validate_homestead_transaction(self, transaction)


class HomesteadTransactionExecutor(FrontierTransactionExecutor):
    pass



from ..arrow_glacier.computation import (
    ArrowGlacierComputation,
)


class GrayGlacierComputation(ArrowGlacierComputation):
    """
    A class for all execution *message* computations in the ``GrayGlacier`` fork.
    Inherits from :class:`~eth.vm.forks.arrow_glacier.ArrowGlacierComputation`
    """

    pass



from eth_utils import (
    ValidationError,
)

from eth._utils.blake2.coders import (
    extract_blake2b_parameters,
)
from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    VMError,
)

try:
    from blake2b import (
        compress as blake2b_compress,
    )
except ImportError:
    from eth._utils.blake2.compression import (
        blake2b_compress,
    )

GAS_COST_PER_ROUND = 1


def blake2b_fcompress(computation: ComputationAPI) -> ComputationAPI:
    try:
        parameters = extract_blake2b_parameters(computation.msg.data_as_bytes)
    except ValidationError as exc:
        raise VMError(f"Blake2b input parameter validation failure: {exc}") from exc

    num_rounds = parameters[0]
    gas_cost = GAS_COST_PER_ROUND * num_rounds

    computation.consume_gas(
        gas_cost, reason=f"Blake2b Compress Precompile w/ {num_rounds} rounds"
    )

    computation.output = blake2b_compress(*parameters)
    return computation



from eth.vm.forks.homestead.state import (
    HomesteadState,
)

from .computation import (
    TangerineWhistleComputation,
)


class TangerineWhistleState(HomesteadState):
    computation_class = TangerineWhistleComputation



from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Tuple,
    Type,
    TypeVar,
)

from eth_utils import (
    to_dict,
    to_set,
)
from eth_utils.toolz import (
    assoc,
    groupby,
)

from eth.abc import (
    ConfigurableAPI,
)


def _is_local_prop(prop: str) -> bool:
    return len(prop.split(".")) == 1


def _extract_top_level_key(prop: str) -> str:
    left, _, _ = prop.partition(".")
    return left


def _extract_tail_key(prop: str) -> str:
    _, _, right = prop.partition(".")
    return right


@to_dict
def _get_local_overrides(overrides: Dict[str, Any]) -> Iterator[Tuple[str, Any]]:
    for prop, value in overrides.items():
        if _is_local_prop(prop):
            yield prop, value


@to_dict
def _get_sub_overrides(overrides: Dict[str, Any]) -> Iterator[Tuple[str, Any]]:
    for prop, value in overrides.items():
        if not _is_local_prop(prop):
            yield prop, value


@to_dict
def _get_sub_overrides_by_prop(
    overrides: Dict[str, Any]
) -> Iterator[Tuple[str, Dict[str, List[str]]]]:
    # we only want the overrides that are not top level.
    sub_overrides = _get_sub_overrides(overrides)
    key_groups = groupby(_extract_top_level_key, sub_overrides.keys())
    for top_level_key, props in key_groups.items():
        yield top_level_key, {
            _extract_tail_key(prop): overrides[prop] for prop in props
        }


@to_set
def _get_top_level_keys(overrides: Dict[str, Any]) -> Iterator[str]:
    for prop in overrides:
        yield _extract_top_level_key(prop)


# Dynamic subclassing is not supported by mypy
# https://github.com/python/mypy/wiki/Unsupported-Python-Features
# Most of the cases where we silence mypy boil down to cases where
# dynamic subclasses where generated through this method


T = TypeVar("T")


class Configurable(ConfigurableAPI):
    """
    Base class for simple inline subclassing
    """

    @classmethod
    def configure(cls: Type[T], __name__: str = None, **overrides: Any) -> Type[T]:
        if __name__ is None:
            __name__ = cls.__name__

        top_level_keys = _get_top_level_keys(overrides)

        # overrides that are *local* to this class.
        local_overrides = _get_local_overrides(overrides)

        for key in top_level_keys:
            if key == "__name__":
                continue
            elif not hasattr(cls, key):
                raise TypeError(
                    f"The {cls.__name__}.configure cannot set attributes that are not "
                    f"already present on the base class. The attribute `{key}` was "
                    f"not found on the base class `{cls}`"
                )

        # overrides that are for sub-properties of this class
        sub_overrides_by_prop = _get_sub_overrides_by_prop(overrides)

        for key, sub_overrides in sub_overrides_by_prop.items():
            sub_cls: Configurable = None
            if key in local_overrides:
                sub_cls = local_overrides[key]
            elif hasattr(cls, key):
                sub_cls = getattr(cls, key)
            else:
                raise Exception(
                    "Invariant: the pre-check that all top level keys are "
                    "present on `cls` should make this code path unreachable"
                )

            if not isinstance(sub_cls, type) or not issubclass(sub_cls, Configurable):
                raise TypeError(
                    f"Unable to configure property `{key}` on class `{cls!r}`.  The "
                    "property being configured must be a subclass of the "
                    "`Configurable` type.  Instead got the following object "
                    f"instance: {sub_cls!r}"
                )

            configured_sub_cls = sub_cls.configure(**sub_overrides)  # type: ignore
            local_overrides = assoc(local_overrides, key, configured_sub_cls)

        return type(__name__, (cls,), local_overrides)



from typing import (
    Any,
    Sequence,
    Type,
)

from cached_property import (
    cached_property,
)
from eth_bloom import (
    BloomFilter,
)
from eth_utils import (
    ValidationError,
    to_bytes,
    to_int,
)
import rlp
from rlp.sedes import (
    Binary,
)

from eth.abc import (
    DecodedZeroOrOneLayerRLP,
    LogAPI,
    ReceiptAPI,
    ReceiptBuilderAPI,
    ReceiptDecoderAPI,
)
from eth.exceptions import (
    UnrecognizedTransactionType,
)
from eth.rlp.receipts import (
    Receipt,
)

from .constants import (
    ACCESS_LIST_TRANSACTION_TYPE,
    VALID_TRANSACTION_TYPES,
)

TYPED_RECEIPT_BODY_CODECS = {
    # Note that the body of a "type 1" receipt uses exactly the same codec as a
    # legacy receipt, so we can simply reuse all the logic.
    ACCESS_LIST_TRANSACTION_TYPE: Receipt,
}


class TypedReceipt(ReceiptAPI, ReceiptDecoderAPI):
    type_id: int
    rlp_type = Binary(min_length=1)  # must have at least one byte for the type
    _inner: ReceiptAPI
    codecs = TYPED_RECEIPT_BODY_CODECS

    def __init__(self, type_id: int, proxy_target: ReceiptAPI) -> None:
        self.type_id = type_id
        self._inner = proxy_target

        # Doing validation here means we don't have to validate on every encode()
        payload_codec = self.get_payload_codec(type_id)
        if not isinstance(proxy_target, payload_codec):
            raise ValidationError(
                f"Cannot embed target {proxy_target!r} "
                f"which doesn't match type ID {type_id}"
            )

    @classmethod
    def decode(cls, encoded: bytes) -> ReceiptAPI:
        type_id = to_int(encoded[0])
        payload = encoded[1:]

        payload_codec = cls.get_payload_codec(type_id)
        inner_receipt = payload_codec.decode(payload)
        return cls(type_id, inner_receipt)

    def encode(self) -> bytes:
        return self._type_byte + self._inner.encode()

    @classmethod
    def get_payload_codec(cls, type_id: int) -> Type[ReceiptDecoderAPI]:
        if type_id in cls.codecs:
            return cls.codecs[type_id]
        elif type_id in VALID_TRANSACTION_TYPES:
            raise UnrecognizedTransactionType(type_id, "Unknown receipt type")
        else:
            raise ValidationError(
                f"Cannot build typed receipt with {hex(type_id)} >= 0x80"
            )

    @classmethod
    def deserialize(cls, encoded_unchecked: DecodedZeroOrOneLayerRLP) -> ReceiptAPI:
        # binary checks a few basics, like the length of the bytes
        encoded = cls.rlp_type.deserialize(encoded_unchecked)
        return cls.decode(encoded)

    @classmethod
    def serialize(cls, obj: "TypedReceipt") -> DecodedZeroOrOneLayerRLP:
        encoded = obj.encode()
        return cls.rlp_type.serialize(encoded)

    @cached_property
    def _type_byte(self) -> bytes:
        return to_bytes(self.type_id)

    @property
    def state_root(self) -> bytes:
        return self._inner.state_root

    @property
    def gas_used(self) -> int:
        return self._inner.gas_used

    @property
    def bloom(self) -> int:
        return self._inner.bloom

    @property
    def logs(self) -> Sequence[LogAPI]:
        return self._inner.logs

    @property
    def bloom_filter(self) -> BloomFilter:
        return self._inner.bloom_filter

    def copy(self, *args: Any, **kwargs: Any) -> ReceiptAPI:
        inner_copy = self._inner.copy(*args, **kwargs)
        return type(self)(self.type_id, inner_copy)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TypedReceipt):
            return False
        else:
            return self.type_id == other.type_id and self._inner == other._inner


class BerlinReceiptBuilder(ReceiptBuilderAPI):
    legacy_sedes = Receipt
    typed_receipt_class = TypedReceipt

    @classmethod
    def decode(cls, encoded: bytes) -> ReceiptAPI:
        if len(encoded) == 0:
            raise ValidationError("Encoded receipt was empty, which makes it invalid")

        type_id = to_int(encoded[0])
        if type_id in cls.typed_receipt_class.codecs:
            return cls.typed_receipt_class.decode(encoded)
        else:
            return rlp.decode(encoded, sedes=cls.legacy_sedes)

    @classmethod
    def deserialize(cls, encoded: DecodedZeroOrOneLayerRLP) -> ReceiptAPI:
        if isinstance(encoded, bytes):
            return cls.typed_receipt_class.deserialize(encoded)
        else:
            return cls.legacy_sedes.deserialize(encoded)

    @classmethod
    def serialize(cls, obj: ReceiptAPI) -> DecodedZeroOrOneLayerRLP:
        return cls.legacy_sedes.serialize(obj)



import logging
from typing import (
    Iterable,
    List,
    Tuple,
    Union,
)

from eth_utils import (
    ValidationError,
    big_endian_to_int,
    int_to_big_endian,
)

from eth.abc import (
    StackAPI,
)
from eth.exceptions import (
    FullStack,
    InsufficientStack,
)
from eth.validation import (
    validate_stack_bytes,
    validate_stack_int,
)


def _busted_type(item_type: type, value: Union[int, bytes]) -> ValidationError:
    return ValidationError(
        f"Stack must always be bytes or int, got {item_type!r} type, val {value!r}"
    )


class Stack(StackAPI):
    """
    VM Stack
    """

    __slots__ = ["values", "_append", "_pop_typed", "__len__"]
    logger = logging.getLogger("eth.vm.stack.Stack")

    #
    # Performance Note: Operations that push to the stack have the data in some natural
    #   form: integer or bytes. Whatever operation is pulling from the stack, also has
    #   its preferred representation to work with. Typically, those two representations
    #   line up (pushed & pulled) so we save a notable amount of conversion time by
    #   storing heterogenous data on the stack, and converting only when necessary.
    #

    def __init__(self) -> None:
        values: List[Tuple[type, Union[int, bytes]]] = []
        self.values = values
        # caching optimizations to avoid an attribute lookup on self.values
        # This doesn't use `cached_property`, because it doesn't play nice with slots
        self._append = values.append
        self._pop_typed = values.pop
        self.__len__ = values.__len__

    def push_int(self, value: int) -> None:
        if len(self.values) > 1023:
            raise FullStack("Stack limit reached")

        validate_stack_int(value)

        self._append((int, value))

    def push_bytes(self, value: bytes) -> None:
        if len(self.values) > 1023:
            raise FullStack("Stack limit reached")

        validate_stack_bytes(value)

        self._append((bytes, value))

    def pop1_bytes(self) -> bytes:
        #
        # Note: This function is optimized for speed over readability.
        # Knowing the popped type means that we can pop *very* quickly
        # when the popped type matches the pushed type.
        #
        if not self.values:
            raise InsufficientStack("Wanted 1 stack item as bytes, had none")
        else:
            item_type, popped = self._pop_typed()
            if item_type is int:
                return int_to_big_endian(popped)  # type: ignore
            elif item_type is bytes:
                return popped  # type: ignore
            else:
                raise _busted_type(item_type, popped)

    def pop1_int(self) -> int:
        #
        # Note: This function is optimized for speed over readability.
        #
        if not self.values:
            raise InsufficientStack("Wanted 1 stack item as int, had none")
        else:
            item_type, popped = self._pop_typed()
            if item_type is int:
                return popped  # type: ignore
            elif item_type is bytes:
                return big_endian_to_int(popped)  # type: ignore
            else:
                raise _busted_type(item_type, popped)

    def pop1_any(self) -> Union[int, bytes]:
        #
        # Note: This function is optimized for speed over readability.
        #
        if not self.values:
            raise InsufficientStack("Wanted 1 stack item, had none")
        else:
            _, popped = self._pop_typed()
            return popped

    def pop_any(self, num_items: int) -> Tuple[Union[int, bytes], ...]:
        #
        # Note: This function is optimized for speed over readability.
        #
        if num_items > len(self.values):
            raise InsufficientStack(
                "Wanted %d stack items, only had %d",
                num_items,
                len(self.values),
            )
        else:
            neg_num_items = -1 * num_items

            # Quickest way to pop off multiple values from the end, in place
            all_popped = reversed(self.values[neg_num_items:])
            del self.values[neg_num_items:]

            # This doesn't use the @to_tuple(generator) pattern, for added performance
            return tuple(val for _, val in all_popped)

    def pop_ints(self, num_items: int) -> Tuple[int, ...]:
        #
        # Note: This function is optimized for speed over readability.
        #
        if num_items > len(self.values):
            raise InsufficientStack(
                "Wanted %d stack items, only had %d",
                num_items,
                len(self.values),
            )
        else:
            neg_num_items = -1 * num_items

            # Quickest way to pop off multiple values from the end, in place
            all_popped = reversed(self.values[neg_num_items:])
            del self.values[neg_num_items:]

            type_cast_popped = []

            # Convert any non-matching types to the requested type (int)
            # This doesn't use the @to_tuple(generator) pattern, for added performance
            for item_type, popped in all_popped:
                if item_type is int:
                    type_cast_popped.append(popped)
                elif item_type is bytes:
                    type_cast_popped.append(big_endian_to_int(popped))  # type: ignore
                else:
                    raise _busted_type(item_type, popped)

            return tuple(type_cast_popped)  # type: ignore

    def pop_bytes(self, num_items: int) -> Tuple[bytes, ...]:
        #
        # Note: This function is optimized for speed over readability.
        #
        if num_items > len(self.values):
            raise InsufficientStack(
                "Wanted %d stack items, only had %d",
                num_items,
                len(self.values),
            )
        else:
            neg_num_items = -1 * num_items

            all_popped = reversed(self.values[neg_num_items:])
            del self.values[neg_num_items:]

            type_cast_popped = []

            # Convert any non-matching types to the requested type (int)
            # This doesn't use the @to_tuple(generator) pattern, for added performance
            for item_type, popped in all_popped:
                if item_type is int:
                    type_cast_popped.append(int_to_big_endian(popped))  # type: ignore
                elif item_type is bytes:
                    type_cast_popped.append(popped)  # type: ignore
                else:
                    raise _busted_type(item_type, popped)

            return tuple(type_cast_popped)

    def swap(self, position: int) -> None:
        idx = -1 * position - 1
        try:
            self.values[-1], self.values[idx] = self.values[idx], self.values[-1]
        except IndexError:
            raise InsufficientStack(f"Insufficient stack items for SWAP{position}")

    def dup(self, position: int) -> None:
        if len(self.values) > 1023:
            raise FullStack("Stack limit reached")

        peek_index = -1 * position
        try:
            self._append(self.values[peek_index])
        except IndexError:
            raise InsufficientStack(f"Insufficient stack items for DUP{position}")

    def _stack_items_str(self) -> Iterable[str]:
        for item_type, val in self.values:
            if isinstance(val, int):
                yield hex(val)
            elif isinstance(val, bytes):
                yield "0x" + val.hex()
            else:
                raise RuntimeError(
                    f"Stack items can only be int or bytes, not {val!r}:{item_type}"
                )

    def __str__(self) -> str:
        return str(list(self._stack_items_str()))



from typing import (
    Type,
)

from eth_hash.auto import (
    keccak,
)
from eth_utils import (
    encode_hex,
)

from eth._utils.address import (
    generate_contract_address,
)
from eth.abc import (
    ComputationAPI,
    MessageAPI,
    SignedTransactionAPI,
    StateAPI,
    TransactionContextAPI,
    TransactionExecutorAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.vm.forks.berlin.state import (
    BerlinState,
    BerlinTransactionExecutor,
)
from eth.vm.message import (
    Message,
)

from .computation import (
    LondonComputation,
)
from .constants import (
    EIP3529_MAX_REFUND_QUOTIENT,
)
from .validation import (
    validate_london_normalized_transaction,
)


class LondonTransactionExecutor(BerlinTransactionExecutor):
    def build_evm_message(self, transaction: SignedTransactionAPI) -> MessageAPI:
        # Use vm_state.get_gas_price instead of transaction_context.gas_price so
        #   that we can run get_transaction_result (aka~ eth_call) and estimate_gas.
        #   Both work better if the GASPRICE opcode returns the original real price,
        #   but the sender's balance doesn't actually deduct the gas. This
        #   get_gas_price() will return 0 for eth_call, but
        #   transaction_context.gas_price will return the same value as the
        #   GASPRICE opcode.
        gas_fee = transaction.gas * self.vm_state.get_gas_price(transaction)

        # Buy Gas
        self.vm_state.delta_balance(transaction.sender, -1 * gas_fee)

        # Increment Nonce
        self.vm_state.increment_nonce(transaction.sender)

        # Setup VM Message
        message_gas = transaction.gas - transaction.intrinsic_gas

        if transaction.to == CREATE_CONTRACT_ADDRESS:
            contract_address = generate_contract_address(
                transaction.sender,
                self.vm_state.get_nonce(transaction.sender) - 1,
            )
            data = b""
            code = transaction.data
        else:
            contract_address = None
            data = transaction.data
            code = self.vm_state.get_code(transaction.to)

        self.vm_state.logger.debug2(
            ("TRANSACTION: %r; sender: %s | to: %s | data-hash: %s"),
            transaction,
            encode_hex(transaction.sender),
            encode_hex(transaction.to),
            encode_hex(keccak(transaction.data)),
        )

        message = Message(
            gas=message_gas,
            to=transaction.to,
            sender=transaction.sender,
            value=transaction.value,
            data=data,
            code=code,
            create_address=contract_address,
        )
        return message

    @classmethod
    def calculate_gas_refund(cls, computation: ComputationAPI, gas_used: int) -> int:
        # Self destruct refunds were added in Frontier
        # London removes them in EIP-3529
        gas_refunded = computation.get_gas_refund()

        return min(gas_refunded, gas_used // EIP3529_MAX_REFUND_QUOTIENT)


class LondonState(BerlinState):
    computation_class = LondonComputation
    transaction_executor_class: Type[TransactionExecutorAPI] = LondonTransactionExecutor

    def get_tip(self, transaction: SignedTransactionAPI) -> int:
        return min(
            transaction.max_fee_per_gas - self.execution_context.base_fee_per_gas,
            transaction.max_priority_fee_per_gas,
        )

    def get_gas_price(self, transaction: SignedTransactionAPI) -> int:
        return min(
            transaction.max_fee_per_gas,
            transaction.max_priority_fee_per_gas
            + self.execution_context.base_fee_per_gas,
        )

    def validate_transaction(self, transaction: SignedTransactionAPI) -> None:
        validate_london_normalized_transaction(
            state=self,
            transaction=transaction,
        )

    def get_transaction_context(
        self: StateAPI, transaction: SignedTransactionAPI
    ) -> TransactionContextAPI:
        """
        London-specific transaction context creation,
        where gas_price includes the block base fee
        """
        effective_gas_price = min(
            transaction.max_priority_fee_per_gas
            + self.execution_context.base_fee_per_gas,
            transaction.max_fee_per_gas,
        )
        # See how this reduces in a pre-1559 transaction:
        # 1. effective_gas_price = min(
        #     transaction.gas_price + self.execution_context.base_fee_per_gas,
        #     transaction.gas_price,
        # )
        # base_fee_per_gas is non-negative, so:
        # 2. effective_gas_price = transaction.gas_price

        return self.get_transaction_context_class()(
            gas_price=effective_gas_price, origin=transaction.sender
        )

    @property
    def base_fee(self: StateAPI) -> int:
        return self.execution_context.base_fee_per_gas



import decimal
import functools
import itertools
from typing import (
    Union,
)

from eth_typing import (
    Hash32,
)
from eth_utils.toolz import (
    curry,
)

from eth.constants import (
    UINT_255_MAX,
    UINT_256_CEILING,
    UINT_256_MAX,
)


def int_to_bytes32(value: Union[int, bool]) -> Hash32:
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValueError(f"Value must be an integer: Got: {type(value)}")
    if value < 0:
        raise ValueError(f"Value cannot be negative: Got: {value}")
    if value > UINT_256_MAX:
        raise ValueError(f"Value exeeds maximum UINT256 size.  Got: {value}")
    value_bytes = value.to_bytes(32, "big")
    return Hash32(value_bytes)


def ceilXX(value: int, ceiling: int) -> int:
    remainder = value % ceiling
    if remainder == 0:
        return value
    else:
        return value + ceiling - remainder


ceil32 = functools.partial(ceilXX, ceiling=32)
ceil8 = functools.partial(ceilXX, ceiling=8)


def unsigned_to_signed(value: int) -> int:
    if value <= UINT_255_MAX:
        return value
    else:
        return value - UINT_256_CEILING


def signed_to_unsigned(value: int) -> int:
    if value < 0:
        return value + UINT_256_CEILING
    else:
        return value


def is_even(value: int) -> bool:
    return value % 2 == 0


def is_odd(value: int) -> bool:
    return value % 2 == 1


def get_highest_bit_index(value: int) -> int:
    value >>= 1
    for bit_length in itertools.count():
        if not value:
            return bit_length
        value >>= 1

    raise Exception("Invariant: unreachable code path")


@curry
def clamp(inclusive_lower_bound: int, inclusive_upper_bound: int, value: int) -> int:
    """
    Bound the given ``value`` between ``inclusive_lower_bound`` and
    ``inclusive_upper_bound``.
    """
    if value <= inclusive_lower_bound:
        return inclusive_lower_bound
    elif value >= inclusive_upper_bound:
        return inclusive_upper_bound
    else:
        return value


def integer_squareroot(value: int) -> int:
    """
    Return the integer square root of ``value``.

    Uses Python's decimal module to compute the square root of ``value`` with
    a precision of 128-bits. The value 128 is chosen since the largest square
    root of a 256-bit integer is a 128-bit integer.
    """
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValueError(f"Value must be an integer: Got: {type(value)}")
    if value < 0:
        raise ValueError(f"Value cannot be negative: Got: {value}")

    with decimal.localcontext() as ctx:
        ctx.prec = 128
        return int(decimal.Decimal(value).sqrt())



from eth_utils import (
    denoms,
)

#
# Difficulty
#
BYZANTIUM_DIFFICULTY_ADJUSTMENT_CUTOFF = 9


EIP649_BLOCK_REWARD = 3 * denoms.ether

EIP658_TRANSACTION_STATUS_CODE_FAILURE = b""
EIP658_TRANSACTION_STATUS_CODE_SUCCESS = b"\x01"



from typing import (
    Any,
    Optional,
)

from toolz import (
    curry,
)

from eth.abc import (
    BlockHeaderAPI,
)
from eth.vm.forks.byzantium.headers import (
    configure_header,
)
from eth.vm.forks.paris.headers import (
    create_paris_header_from_parent,
)

from .blocks import (
    ShanghaiBlockHeader,
)


@curry
def create_shanghai_header_from_parent(
    parent_header: Optional[BlockHeaderAPI],
    **header_params: Any,
) -> BlockHeaderAPI:
    paris_validated_header = create_paris_header_from_parent(
        parent_header, **header_params
    )

    # extract params validated up to paris (previous VM)
    # and plug into a `ShanghaiBlockHeader` class
    all_fields = paris_validated_header.as_dict()
    return ShanghaiBlockHeader(**all_fields)


configure_shanghai_header = configure_header()



from .builders import (  # noqa: F401
    at_block_number,
    build,
    chain_id,
    chain_split,
    copy,
    dao_fork_at,
    disable_dao_fork,
    disable_pow_check,
    enable_pow_mining,
    fork_at,
    genesis,
    import_block,
    import_blocks,
    mine_block,
    mine_blocks,
    name,
)
from .builders import (  # noqa: F401
    byzantium_at,
    frontier_at,
    homestead_at,
    spurious_dragon_at,
    tangerine_whistle_at,
    constantinople_at,
    petersburg_at,
    istanbul_at,
    muir_glacier_at,
    berlin_at,
    london_at,
    arrow_glacier_at,
    gray_glacier_at,
    paris_at,
    shanghai_at,
    latest_mainnet_at,
)


mining_mainnet_fork_at_fns = (
    byzantium_at,
    frontier_at,
    homestead_at,
    spurious_dragon_at,
    tangerine_whistle_at,
    petersburg_at,
    istanbul_at,
    muir_glacier_at,
    berlin_at,
    london_at,
    arrow_glacier_at,
    gray_glacier_at,
)
pos_mainnet_fork_at_fns = (
    paris_at,
    shanghai_at,
)
mainnet_fork_at_fns = mining_mainnet_fork_at_fns + pos_mainnet_fork_at_fns


class API:
    #
    # Chain Class Construction
    #

    # Primary wrapper function
    build = staticmethod(build)

    # Configure chain vm_configuration
    fork_at = staticmethod(fork_at)

    # Configure chain name
    name = staticmethod(name)

    # Configure chain chain_id
    chain_id = staticmethod(chain_id)

    # Mainnet Forks
    frontier_at = staticmethod(frontier_at)
    homestead_at = staticmethod(homestead_at)
    tangerine_whistle_at = staticmethod(tangerine_whistle_at)
    spurious_dragon_at = staticmethod(spurious_dragon_at)
    byzantium_at = staticmethod(byzantium_at)
    constantinople_at = staticmethod(constantinople_at)
    istanbul_at = staticmethod(istanbul_at)
    muir_glacier_at = staticmethod(muir_glacier_at)
    berlin_at = staticmethod(berlin_at)
    london_at = staticmethod(london_at)
    arrow_glacier_at = staticmethod(arrow_glacier_at)
    gray_glacier_at = staticmethod(gray_glacier_at)
    paris_at = staticmethod(paris_at)
    shanghai_at = staticmethod(shanghai_at)

    # iterable of the fork specific functions
    mainnet_fork_at_fns = mainnet_fork_at_fns
    mining_mainnet_fork_at_fns = mining_mainnet_fork_at_fns

    # DAO Fork specific
    dao_fork_at = staticmethod(dao_fork_at)
    disable_dao_fork = staticmethod(disable_dao_fork)

    # Chain Mining config
    enable_pow_mining = staticmethod(enable_pow_mining)
    disable_pow_check = staticmethod(disable_pow_check)

    #
    # Chain Instance Initialization
    #
    genesis = staticmethod(genesis)

    #
    # Chain Building
    #
    copy = staticmethod(copy)

    import_block = staticmethod(import_block)
    import_blocks = staticmethod(import_blocks)

    mine_block = staticmethod(mine_block)
    mine_blocks = staticmethod(mine_blocks)

    chain_split = staticmethod(chain_split)
    at_block_number = staticmethod(at_block_number)


api = API()



import functools

from eth.abc import (
    ComputationAPI,
)


def dup_XX(computation: ComputationAPI, position: int) -> None:
    """
    Stack item duplication.
    """
    computation.stack_dup(position)


dup1 = functools.partial(dup_XX, position=1)
dup2 = functools.partial(dup_XX, position=2)
dup3 = functools.partial(dup_XX, position=3)
dup4 = functools.partial(dup_XX, position=4)
dup5 = functools.partial(dup_XX, position=5)
dup6 = functools.partial(dup_XX, position=6)
dup7 = functools.partial(dup_XX, position=7)
dup8 = functools.partial(dup_XX, position=8)
dup9 = functools.partial(dup_XX, position=9)
dup10 = functools.partial(dup_XX, position=10)
dup11 = functools.partial(dup_XX, position=11)
dup12 = functools.partial(dup_XX, position=12)
dup13 = functools.partial(dup_XX, position=13)
dup14 = functools.partial(dup_XX, position=14)
dup15 = functools.partial(dup_XX, position=15)
dup16 = functools.partial(dup_XX, position=16)



from typing import (
    NamedTuple,
)

from eth_keys import (
    datatypes,
    keys,
)
from eth_keys.exceptions import (
    BadSignature,
)
from eth_utils import (
    ValidationError,
    int_to_big_endian,
)
import rlp

from eth._utils.numeric import (
    is_even,
)
from eth.abc import (
    SignedTransactionAPI,
    UnsignedTransactionAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.rlp.transactions import (
    BaseTransaction,
)
from eth.typing import (
    VRS,
    Address,
)

EIP155_CHAIN_ID_OFFSET = 35
# Add this offset to y_parity to get "v" for legacy transactions, from Frontier
V_OFFSET = 27


def is_eip_155_signed_transaction(transaction: BaseTransaction) -> bool:
    return transaction.v >= EIP155_CHAIN_ID_OFFSET


def extract_chain_id(v: int) -> int:
    if is_even(v):
        return (v - EIP155_CHAIN_ID_OFFSET - 1) // 2
    else:
        return (v - EIP155_CHAIN_ID_OFFSET) // 2


def extract_signature_v(v: int) -> int:
    if is_even(v):
        return V_OFFSET + 1
    else:
        return V_OFFSET


def create_transaction_signature(
    unsigned_txn: UnsignedTransactionAPI,
    private_key: datatypes.PrivateKey,
    chain_id: int = None,
) -> VRS:
    transaction_parts = rlp.decode(rlp.encode(unsigned_txn))

    if chain_id:
        transaction_parts_for_signature = transaction_parts + [
            int_to_big_endian(chain_id),
            b"",
            b"",
        ]
    else:
        transaction_parts_for_signature = transaction_parts

    message = rlp.encode(transaction_parts_for_signature)
    signature = private_key.sign_msg(message)

    canonical_v, r, s = signature.vrs

    if chain_id:
        v = canonical_v + chain_id * 2 + EIP155_CHAIN_ID_OFFSET
    else:
        v = canonical_v + V_OFFSET

    return VRS((v, r, s))


def validate_transaction_signature(transaction: SignedTransactionAPI) -> None:
    message = transaction.get_message_for_signing()
    vrs = (transaction.y_parity, transaction.r, transaction.s)
    try:
        signature = keys.Signature(vrs=vrs)
        public_key = signature.recover_public_key_from_msg(message)
    except BadSignature as e:
        raise ValidationError(f"Bad Signature: {str(e)}")

    if not signature.verify_msg(message, public_key):
        raise ValidationError("Invalid Signature")


def extract_transaction_sender(transaction: SignedTransactionAPI) -> Address:
    vrs = (transaction.y_parity, transaction.r, transaction.s)
    signature = keys.Signature(vrs=vrs)
    message = transaction.get_message_for_signing()
    public_key = signature.recover_public_key_from_msg(message)
    sender = public_key.to_canonical_address()
    return Address(sender)


class IntrinsicGasSchedule(NamedTuple):
    gas_tx: int
    gas_txcreate: int
    gas_txdatazero: int
    gas_txdatanonzero: int


def calculate_intrinsic_gas(
    gas_schedule: IntrinsicGasSchedule,
    transaction: SignedTransactionAPI,
) -> int:
    num_zero_bytes = transaction.data.count(b"\x00")
    num_non_zero_bytes = len(transaction.data) - num_zero_bytes
    if transaction.to == CREATE_CONTRACT_ADDRESS:
        create_cost = gas_schedule.gas_txcreate
    else:
        create_cost = 0
    return (
        gas_schedule.gas_tx
        + num_zero_bytes * gas_schedule.gas_txdatazero
        + num_non_zero_bytes * gas_schedule.gas_txdatanonzero
        + create_cost
    )



import logging

from eth_typing import (
    Address,
)

from eth.abc import (
    MessageAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.typing import (
    BytesOrView,
)
from eth.validation import (
    validate_canonical_address,
    validate_gte,
    validate_is_boolean,
    validate_is_bytes,
    validate_is_bytes_or_view,
    validate_is_integer,
    validate_uint256,
)


class Message(MessageAPI):
    __slots__ = [
        "to",
        "sender",
        "value",
        "data",
        "depth",
        "gas",
        "code",
        "_code_address",
        "create_address",
        "should_transfer_value",
        "is_static",
        "_storage_address",
    ]

    logger = logging.getLogger("eth.vm.message.Message")

    def __init__(
        self,
        gas: int,
        to: Address,
        sender: Address,
        value: int,
        data: BytesOrView,
        code: bytes,
        depth: int = 0,
        create_address: Address = None,
        code_address: Address = None,
        should_transfer_value: bool = True,
        is_static: bool = False,
    ) -> None:
        validate_uint256(gas, title="Message.gas")
        self.gas: int = gas

        if to != CREATE_CONTRACT_ADDRESS:
            validate_canonical_address(to, title="Message.to")
        self.to = to

        validate_canonical_address(sender, title="Message.sender")
        self.sender = sender

        validate_uint256(value, title="Message.value")
        self.value = value

        validate_is_bytes_or_view(data, title="Message.data")
        self.data = data

        validate_is_integer(depth, title="Message.depth")
        validate_gte(depth, minimum=0, title="Message.depth")
        self.depth = depth

        validate_is_bytes(code, title="Message.code")
        self.code = code

        if create_address is not None:
            validate_canonical_address(create_address, title="Message.storage_address")
        self.storage_address = create_address

        if code_address is not None:
            validate_canonical_address(code_address, title="Message.code_address")
        self.code_address = code_address

        validate_is_boolean(
            should_transfer_value, title="Message.should_transfer_value"
        )
        self.should_transfer_value = should_transfer_value

        validate_is_boolean(is_static, title="Message.is_static")
        self.is_static = is_static

    @property
    def code_address(self) -> Address:
        if self._code_address is not None:
            return self._code_address
        else:
            return self.to

    @code_address.setter
    def code_address(self, value: Address) -> None:
        self._code_address = value

    @property
    def storage_address(self) -> Address:
        if self._storage_address is not None:
            return self._storage_address
        else:
            return self.to

    @storage_address.setter
    def storage_address(self, value: Address) -> None:
        self._storage_address = value

    @property
    def is_create(self) -> bool:
        return self.to == CREATE_CONTRACT_ADDRESS

    @property
    def data_as_bytes(self) -> bytes:
        return bytes(self.data)



from typing import (
    TYPE_CHECKING,
    Any,
)

from eth._utils.db import (
    get_parent_header,
)
from eth._utils.headers import (
    compute_gas_limit,
    fill_header_params_from_parent,
    new_timestamp_from_parent,
)
from eth.abc import (
    BlockHeaderAPI,
)
from eth.constants import (
    BOMB_EXPONENTIAL_FREE_PERIODS,
    BOMB_EXPONENTIAL_PERIOD,
    DIFFICULTY_ADJUSTMENT_DENOMINATOR,
    DIFFICULTY_MINIMUM,
    GENESIS_GAS_LIMIT,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.validation import (
    validate_gt,
    validate_header_params_for_configuration,
)

from .constants import (
    FRONTIER_DIFFICULTY_ADJUSTMENT_CUTOFF,
)

if TYPE_CHECKING:
    from eth.vm.forks.frontier import FrontierVM  # noqa: F401


def compute_frontier_difficulty(parent_header: BlockHeaderAPI, timestamp: int) -> int:
    """
    Computes the difficulty for a frontier block based on the parent block.
    """
    validate_gt(timestamp, parent_header.timestamp, title="Header timestamp")

    offset = parent_header.difficulty // DIFFICULTY_ADJUSTMENT_DENOMINATOR

    # We set the minimum to the lowest of the protocol minimum and the parent
    # minimum to allow for the initial frontier *warming* period during which
    # the difficulty begins lower than the protocol minimum.
    difficulty_minimum = min(parent_header.difficulty, DIFFICULTY_MINIMUM)

    if timestamp - parent_header.timestamp < FRONTIER_DIFFICULTY_ADJUSTMENT_CUTOFF:
        base_difficulty = max(
            parent_header.difficulty + offset,
            difficulty_minimum,
        )
    else:
        base_difficulty = max(
            parent_header.difficulty - offset,
            difficulty_minimum,
        )

    # Adjust for difficulty bomb.
    num_bomb_periods = (
        (parent_header.block_number + 1) // BOMB_EXPONENTIAL_PERIOD
    ) - BOMB_EXPONENTIAL_FREE_PERIODS

    if num_bomb_periods >= 0:
        difficulty = max(
            base_difficulty + 2**num_bomb_periods,
            DIFFICULTY_MINIMUM,
        )
    else:
        difficulty = base_difficulty

    return difficulty


def create_frontier_header_from_parent(
    parent_header: BlockHeaderAPI, **header_params: Any
) -> BlockHeader:
    if "timestamp" not in header_params:
        header_params["timestamp"] = new_timestamp_from_parent(parent_header)

    if "difficulty" not in header_params:
        # Use setdefault to ensure the new header has the same timestamp we use to
        # calculate its difficulty.
        header_params["difficulty"] = compute_frontier_difficulty(
            parent_header,
            header_params["timestamp"],
        )
    if "gas_limit" not in header_params:
        header_params["gas_limit"] = compute_gas_limit(
            parent_header,
            genesis_gas_limit=GENESIS_GAS_LIMIT,
        )

    all_fields = fill_header_params_from_parent(parent_header, **header_params)
    return BlockHeader(**all_fields)


def configure_frontier_header(vm: "FrontierVM", **header_params: Any) -> BlockHeader:
    validate_header_params_for_configuration(header_params)

    with vm.get_header().build_changeset(**header_params) as changeset:
        if "timestamp" in header_params and vm.get_header().block_number > 0:
            parent_header = get_parent_header(changeset.build_rlp(), vm.chaindb)
            changeset.difficulty = compute_frontier_difficulty(
                parent_header,
                header_params["timestamp"],
            )

        header = changeset.commit()
    return header



from typing import (
    Iterable,
)

from eth_typing import (
    Address,
)
from eth_utils import (
    to_set,
)

from eth import (
    constants,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth.abc import (
    ComputationAPI,
)

THREE = force_bytes_to_address(b"\x03")


@to_set
def collect_touched_accounts(
    computation: ComputationAPI, ancestor_had_error: bool = False
) -> Iterable[Address]:
    """
    Collect all of the accounts that *may* need to be deleted based on
    `EIP-161 <https://eips.ethereum.org/EIPS/eip-161>`_.

    Checking whether they *do* need to be deleted happens in the caller.

    See also: https://github.com/ethereum/EIPs/issues/716
    """
    # EIP-161:
    # The coinbase is always touched via block transaction fee and block rewards
    # (pre-merge).
    yield computation.state.coinbase

    # collect those explicitly marked for deletion ("beneficiary" is of SELFDESTRUCT)
    for beneficiary in sorted(set(computation.accounts_to_delete.values())):
        if computation.is_error or ancestor_had_error:
            # Special case to account for geth+parity bug
            # https://github.com/ethereum/EIPs/issues/716
            if beneficiary == THREE:
                yield beneficiary
            continue
        else:
            yield beneficiary

    # collect account directly addressed
    if computation.msg.to != constants.CREATE_CONTRACT_ADDRESS:
        if computation.is_error or ancestor_had_error:
            # collect RIPEMD160 precompile even if ancestor computation had error;
            # otherwise, skip collection from children of errored-out computations;
            # if there were no special-casing for RIPEMD160, we'd simply `pass` here
            if computation.msg.to == THREE:
                yield computation.msg.to
        else:
            yield computation.msg.to

    # recurse into nested computations (even errored ones, since looking for RIPEMD160)
    for child in computation.children:
        yield from collect_touched_accounts(
            child, ancestor_had_error=(computation.is_error or ancestor_had_error)
        )



import functools
from typing import (
    Tuple,
)

from eth import (
    constants,
)
from eth.abc import (
    ComputationAPI,
)


def log_XX(computation: ComputationAPI, topic_count: int) -> None:
    if topic_count < 0 or topic_count > 4:
        raise TypeError("Invalid log topic size.  Must be 0, 1, 2, 3, or 4")

    mem_start_position, size = computation.stack_pop_ints(2)

    if not topic_count:
        topics: Tuple[int, ...] = ()
    elif topic_count > 1:
        topics = computation.stack_pop_ints(topic_count)
    else:
        topics = (computation.stack_pop1_int(),)

    data_gas_cost = constants.GAS_LOGDATA * size
    topic_gas_cost = constants.GAS_LOGTOPIC * topic_count
    total_gas_cost = data_gas_cost + topic_gas_cost

    computation.consume_gas(
        total_gas_cost,
        reason="Log topic and data gas cost",
    )

    computation.extend_memory(mem_start_position, size)
    log_data = computation.memory_read_bytes(mem_start_position, size)

    computation.add_log_entry(
        account=computation.msg.storage_address,
        topics=topics,
        data=log_data,
    )


log0 = functools.partial(log_XX, topic_count=0)
log1 = functools.partial(log_XX, topic_count=1)
log2 = functools.partial(log_XX, topic_count=2)
log3 = functools.partial(log_XX, topic_count=3)
log4 = functools.partial(log_XX, topic_count=4)



from typing import Tuple, Type
from eth_typing import BlockNumber
from eth_utils import decode_hex

from .constants import (
    BERLIN_ROPSTEN_BLOCK,
    BYZANTIUM_ROPSTEN_BLOCK,
    CONSTANTINOPLE_ROPSTEN_BLOCK,
    ISTANBUL_ROPSTEN_BLOCK,
    MUIR_GLACIER_ROPSTEN_BLOCK,
    PETERSBURG_ROPSTEN_BLOCK,
    ROPSTEN_CHAIN_ID,
    SPURIOUS_DRAGON_ROPSTEN_BLOCK,
    TANGERINE_WHISTLE_ROPSTEN_BLOCK,
)
from eth import constants

from eth.abc import VirtualMachineAPI
from eth.chains.base import Chain
from eth.rlp.headers import BlockHeader
from eth.vm.forks import (
    BerlinVM,
    ByzantiumVM,
    ConstantinopleVM,
    IstanbulVM,
    MuirGlacierVM,
    PetersburgVM,
    SpuriousDragonVM,
    TangerineWhistleVM,
)


ROPSTEN_VM_CONFIGURATION = (
    # Note: Frontier and Homestead are excluded since this chain
    # starts at Tangerine Whistle.
    (TANGERINE_WHISTLE_ROPSTEN_BLOCK, TangerineWhistleVM),
    (SPURIOUS_DRAGON_ROPSTEN_BLOCK, SpuriousDragonVM),
    (BYZANTIUM_ROPSTEN_BLOCK, ByzantiumVM),
    (CONSTANTINOPLE_ROPSTEN_BLOCK, ConstantinopleVM),
    (PETERSBURG_ROPSTEN_BLOCK, PetersburgVM),
    (ISTANBUL_ROPSTEN_BLOCK, IstanbulVM),
    (MUIR_GLACIER_ROPSTEN_BLOCK, MuirGlacierVM),
    (BERLIN_ROPSTEN_BLOCK, BerlinVM),
)


class BaseRopstenChain:
    vm_configuration: Tuple[
        Tuple[BlockNumber, Type[VirtualMachineAPI]], ...
    ] = ROPSTEN_VM_CONFIGURATION
    chain_id: int = ROPSTEN_CHAIN_ID


class RopstenChain(BaseRopstenChain, Chain):
    pass


ROPSTEN_GENESIS_HEADER = BlockHeader(
    difficulty=1048576,
    extra_data=decode_hex(
        "0x3535353535353535353535353535353535353535353535353535353535353535"
    ),
    gas_limit=16777216,
    gas_used=0,
    bloom=0,
    mix_hash=constants.ZERO_HASH32,
    nonce=constants.GENESIS_NONCE,
    block_number=constants.GENESIS_BLOCK_NUMBER,
    parent_hash=constants.ZERO_HASH32,
    receipt_root=constants.BLANK_ROOT_HASH,
    uncles_hash=constants.EMPTY_UNCLE_HASH,
    state_root=decode_hex(
        "0x217b0bbcfb72e2d57e28f33cb361b9983513177755dc3f33ce3e7022ed62b77b"
    ),
    timestamp=0,
    transaction_root=constants.BLANK_ROOT_HASH,
)



import copy
import random
from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Type,
)

from eth_keys import (
    keys,
)
from eth_typing import (
    Address,
)
from eth_utils import (
    int_to_big_endian,
)

from eth._utils.db import (
    apply_state_dict,
)
from eth._utils.padding import (
    pad32,
)
from eth.constants import (
    BLANK_ROOT_HASH,
)
from eth.db.atomic import (
    AtomicDB,
)
from eth.typing import (
    AccountState,
    TransactionDict,
)
from eth.vm.state import (
    BaseState,
)


def wrap_in_list(item: Any) -> List[Any]:
    return [item]


def add_transaction_to_group(
    group: Dict[str, Any], transaction: TransactionDict
) -> Tuple[Dict[str, Any], Dict[str, int]]:
    for key in ["gasPrice", "nonce", "secretKey", "to"]:
        if key in transaction and transaction[key] != group[key]:  # type: ignore # https://github.com/python/mypy/issues/5359 # noqa: 501
            raise ValueError(f"Can't add transaction as it differs in {key}")

    new_group = copy.deepcopy(group)
    indexes = {}
    for key, index_key in [("data", "data"), ("gasLimit", "gas"), ("value", "value")]:
        if key in group:
            if key not in transaction:
                if len(new_group[key]) != 1:
                    raise ValueError(f"Can't add transaction as {key} is ambiguous")
                index = 0
            else:
                if transaction[key] not in new_group[key]:  # type: ignore # https://github.com/python/mypy/issues/5359 # noqa: 501
                    new_group[key].append(transaction[key])  # type: ignore # https://github.com/python/mypy/issues/5359 # noqa: 501
                index = new_group[key].index(transaction[key])  # type: ignore # https://github.com/python/mypy/issues/5359 # noqa: 501
            indexes[index_key] = index
        else:
            assert key not in transaction
    return new_group, indexes


def calc_state_root(state_dict: AccountState, state_class: Type[BaseState]) -> bytes:
    state = state_class(AtomicDB(), None, BLANK_ROOT_HASH)
    apply_state_dict(state, state_dict)
    return state.state_root


def generate_random_keypair() -> Tuple[bytes, Address]:
    key_object = keys.PrivateKey(pad32(int_to_big_endian(random.getrandbits(8 * 32))))
    return key_object.to_bytes(), Address(key_object.public_key.to_canonical_address())


def generate_random_address() -> Address:
    _, address = generate_random_keypair()
    return address



from ..london.computation import (
    LondonComputation,
)


class ArrowGlacierComputation(LondonComputation):
    """
    A class for all execution *message* computations in the ``ArrowGlacier`` fork.
    Inherits from :class:`~eth.vm.forks.london.LondonComputation`
    """

    pass



GAS_MOD_EXP_QUADRATIC_DENOMINATOR_EIP_2565 = 3

# EIP 2718
VALID_TRANSACTION_TYPES = range(0, 0x80)

# EIP 2929
WARM_STORAGE_READ_COST = 100
COLD_SLOAD_COST = 2100
COLD_ACCOUNT_ACCESS_COST = 2600

# EIP 2930
ACCESS_LIST_TRANSACTION_TYPE = 1
ACCESS_LIST_ADDRESS_COST_EIP_2930 = 2400
ACCESS_LIST_STORAGE_KEY_COST_EIP_2930 = 1900



from eth_utils import (
    ValidationError,
)

from eth.abc import (
    BlockHeaderAPI,
    SignedTransactionAPI,
    StateAPI,
    VirtualMachineAPI,
)


def validate_frontier_transaction(
    state: StateAPI, transaction: SignedTransactionAPI
) -> None:
    max_gas_cost = transaction.gas * state.get_gas_price(transaction)
    sender_balance = state.get_balance(transaction.sender)

    if sender_balance < max_gas_cost:
        raise ValidationError(
            f"Sender {transaction.sender!r} cannot afford txn gas "
            f"{max_gas_cost} with account balance {sender_balance}"
        )

    total_cost = transaction.value + max_gas_cost

    if sender_balance < total_cost:
        raise ValidationError(
            f"Sender does not have enough balance to cover transaction value and gas "
            f" (has {sender_balance}, needs {total_cost})"
        )

    sender_nonce = state.get_nonce(transaction.sender)
    if sender_nonce != transaction.nonce:
        raise ValidationError(
            f"Invalid transaction nonce: Expected {sender_nonce}, "
            f"but got {transaction.nonce}"
        )


def validate_frontier_transaction_against_header(
    _vm: VirtualMachineAPI,
    base_header: BlockHeaderAPI,
    transaction: SignedTransactionAPI,
) -> None:
    if base_header.gas_used + transaction.gas > base_header.gas_limit:
        raise ValidationError(
            f"Transaction exceeds gas limit: using {transaction.gas}, "
            f"bringing total to {base_header.gas_used + transaction.gas}, "
            f"but limit is {base_header.gas_limit}"
        )



from eth_utils import decode_hex

from .constants import (
    BERLIN_GOERLI_BLOCK,
    ISTANBUL_GOERLI_BLOCK,
    PETERSBURG_GOERLI_BLOCK,
)

from eth import constants

from eth.rlp.headers import BlockHeader
from eth.vm.forks import (
    BerlinVM,
    IstanbulVM,
    PetersburgVM,
)

GOERLI_VM_CONFIGURATION = (
    (PETERSBURG_GOERLI_BLOCK, PetersburgVM),
    (ISTANBUL_GOERLI_BLOCK, IstanbulVM),
    (BERLIN_GOERLI_BLOCK, BerlinVM),
)


GOERLI_GENESIS_HEADER = BlockHeader(
    block_number=0,
    bloom=0,
    coinbase=constants.ZERO_ADDRESS,
    difficulty=1,
    extra_data=decode_hex(
        "0x22466c6578692069732061207468696e6722202d204166726900000000000000e0a2bd4258d2768837baa26a28fe71dc079f84c70000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"  # noqa: E501
    ),
    gas_limit=10485760,
    gas_used=0,
    mix_hash=constants.ZERO_HASH32,
    nonce=decode_hex("0x0000000000000000"),
    parent_hash=constants.GENESIS_PARENT_HASH,
    receipt_root=constants.BLANK_ROOT_HASH,
    state_root=decode_hex(
        "0x5d6cded585e73c4e322c30c2f782a336316f17dd85a4863b9d838d2d4b8b3008"
    ),
    timestamp=1548854791,
    transaction_root=constants.BLANK_ROOT_HASH,
    uncles_hash=constants.EMPTY_UNCLE_HASH,
)



from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.state import BaseState

from .blocks import ArrowGlacierBlock
from .headers import (
    compute_arrow_glacier_difficulty,
    configure_arrow_glacier_header,
    create_arrow_glacier_header_from_parent,
)
from .state import ArrowGlacierState
from .. import LondonVM


class ArrowGlacierVM(LondonVM):
    # fork name
    fork = "arrow-glacier"

    # classes
    block_class: Type[BaseBlock] = ArrowGlacierBlock
    _state_class: Type[BaseState] = ArrowGlacierState

    # Methods
    create_header_from_parent = staticmethod(  # type: ignore
        create_arrow_glacier_header_from_parent(compute_arrow_glacier_difficulty)
    )
    compute_difficulty = staticmethod(compute_arrow_glacier_difficulty)  # type: ignore
    configure_header = configure_arrow_glacier_header



import functools
import itertools
from typing import (
    Dict,
    Iterable,
    Sequence,
    Tuple,
    Type,
    cast,
)

from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    BlockNumber,
    Hash32,
)
from eth_utils import (
    encode_hex,
)
from trie.exceptions import (
    MissingTrieNode,
)
from web3 import Web3

from eth._warnings import (
    catch_and_ignore_import_warning,
)
from eth.abc import (
    AtomicDatabaseAPI,
    BlockAPI,
    BlockHeaderAPI,
    ChainDatabaseAPI,
    DatabaseAPI,
    ReceiptAPI,
    ReceiptDecoderAPI,
    SignedTransactionAPI,
    TransactionDecoderAPI,
    WithdrawalAPI,
)
from eth.constants import (
    EMPTY_UNCLE_HASH,
    GENESIS_PARENT_HASH,
)
from eth.db.chain_gaps import (
    GENESIS_CHAIN_GAPS,
    GapChange,
    GapInfo,
    fill_gap,
    is_block_number_in_gap,
    reopen_gap,
)
from eth.db.header import (
    HeaderDB,
)
from eth.db.schema import (
    SchemaV1,
)
from eth.db.trie import (
    make_trie_root_and_nodes,
)
from eth.exceptions import (
    HeaderNotFound,
    ReceiptNotFound,
    TransactionNotFound,
)
from eth.rlp.sedes import (
    chain_gaps,
)
from eth.typing import (
    ChainGaps,
)
from eth.validation import (
    validate_word,
)
from eth.vm.forks.shanghai.withdrawals import (
    Withdrawal,
)
from eth.vm.header import (
    HeaderSedes,
)

with catch_and_ignore_import_warning():
    from eth_utils import (
        ValidationError,
        to_tuple,
    )
    import rlp
    from trie import (
        HexaryTrie,
    )


class BlockDataKey(rlp.Serializable):
    # used for transactions and withdrawals
    fields = [
        ("block_number", rlp.sedes.big_endian_int),
        ("index", rlp.sedes.big_endian_int),
    ]


class ChainDB(HeaderDB, ChainDatabaseAPI):
    w3: Web3

    def __init__(self, w3: Web3) -> None:
        self.w3 = w3

    def get_chain_gaps(self) -> ChainGaps:
        return self._get_chain_gaps(self.db)

    @classmethod
    def _get_chain_gaps(cls, db: DatabaseAPI) -> ChainGaps:
        try:
            encoded_gaps = db[SchemaV1.make_chain_gaps_lookup_key()]
        except KeyError:
            return GENESIS_CHAIN_GAPS
        else:
            return rlp.decode(encoded_gaps, sedes=chain_gaps)

    @classmethod
    def _update_chain_gaps(
        cls, db: DatabaseAPI, persisted_block: BlockAPI, base_gaps: ChainGaps = None
    ) -> GapInfo:
        # If we make many updates in a row, we can avoid reloading the integrity info by
        # continuously caching it and providing it as a parameter to this API
        if base_gaps is None:
            base_gaps = cls._get_chain_gaps(db)

        gap_change, gaps = fill_gap(persisted_block.number, base_gaps)
        if gap_change is not GapChange.NoChange:
            db.set(
                SchemaV1.make_chain_gaps_lookup_key(),
                rlp.encode(gaps, sedes=chain_gaps),
            )

        return gap_change, gaps

    @classmethod
    def _update_header_chain_gaps(
        cls,
        db: DatabaseAPI,
        persisting_header: BlockHeaderAPI,
        base_gaps: ChainGaps = None,
    ) -> GapInfo:
        # The only reason we overwrite this here is to be able to detect when the
        # HeaderDB de-canonicalizes an uncle that should cause us to
        # re-open a block gap.
        gap_change, gaps = super()._update_header_chain_gaps(
            db, persisting_header, base_gaps
        )

        if gap_change is not GapChange.NoChange or persisting_header.block_number == 0:
            return gap_change, gaps

        # We have written a header for which block number we've already had a header.
        # This might be a sign of a de-canonicalized uncle.
        current_gaps = cls._get_chain_gaps(db)
        if not is_block_number_in_gap(persisting_header.block_number, current_gaps):
            # ChainDB believes we have that block. If the header has changed, we need to
            # re-open a gap for the corresponding block.
            old_canonical_header = cls._get_canonical_block_header_by_number(
                db, persisting_header.block_number
            )
            if old_canonical_header != persisting_header:
                updated_gaps = reopen_gap(persisting_header.block_number, current_gaps)
                db.set(
                    SchemaV1.make_chain_gaps_lookup_key(),
                    rlp.encode(updated_gaps, sedes=chain_gaps),
                )

        return gap_change, gaps

    #
    # Header API
    #
    def get_block_uncles(self, uncles_hash: Hash32) -> Tuple[BlockHeaderAPI, ...]:
        validate_word(uncles_hash, title="Uncles Hash")
        if uncles_hash == EMPTY_UNCLE_HASH:
            return ()
        try:
            encoded_uncles = self.db[uncles_hash]
        except KeyError as exc:
            raise HeaderNotFound(f"No uncles found for hash {uncles_hash!r}") from exc
        else:
            return tuple(
                rlp.decode(encoded_uncles, sedes=rlp.sedes.CountableList(HeaderSedes))
            )

    @classmethod
    def _decanonicalize_old_headers(
        cls,
        db: DatabaseAPI,
        numbers_to_decanonicalize: Sequence[BlockNumber],
    ) -> Tuple[BlockHeaderAPI, ...]:
        old_canonical_headers = []

        # remove transaction lookups for blocks that are no longer canonical
        for block_number in numbers_to_decanonicalize:
            try:
                old_hash = cls._get_canonical_block_hash(db, block_number)
            except HeaderNotFound:
                # no old block, and no more possible
                break
            else:
                old_header = cls._get_block_header_by_hash(db, old_hash)
                old_canonical_headers.append(old_header)
                try:
                    transaction_hashes = cls._get_block_transaction_hashes(
                        db, old_header
                    )
                    for transaction_hash in transaction_hashes:
                        cls._remove_transaction_from_canonical_chain(
                            db, transaction_hash
                        )
                except MissingTrieNode:
                    # If the transactions were never stored for the (now) non-canonical
                    # chain, then you don't need to remove them from the canonical chain
                    # lookup.
                    pass

        return tuple(old_canonical_headers)

    #
    # Block API
    #
    def persist_block(
        self, block: BlockAPI, genesis_parent_hash: Hash32 = GENESIS_PARENT_HASH
    ) -> Tuple[Tuple[Hash32, ...], Tuple[Hash32, ...]]:
        with self.db.atomic_batch() as db:
            return self._persist_block(db, block, genesis_parent_hash)

    def persist_unexecuted_block(
        self,
        block: BlockAPI,
        receipts: Tuple[ReceiptAPI, ...],
        genesis_parent_hash: Hash32 = GENESIS_PARENT_HASH,
    ) -> Tuple[Tuple[Hash32, ...], Tuple[Hash32, ...]]:
        tx_root_hash, tx_kv_nodes = make_trie_root_and_nodes(block.transactions)

        if tx_root_hash != block.header.transaction_root:
            raise ValidationError(
                f"Block's transaction_root ({block.header.transaction_root!r}) "
                f"does not match expected value: {tx_root_hash!r}"
            )

        receipt_root_hash, receipt_kv_nodes = make_trie_root_and_nodes(receipts)

        if receipt_root_hash != block.header.receipt_root:
            raise ValidationError(
                f"Block's receipt_root ({block.header.receipt_root!r}) "
                f"does not match expected value: {receipt_root_hash!r}"
            )

        with self.db.atomic_batch() as db:
            self._persist_trie_data_dict(db, receipt_kv_nodes)
            self._persist_trie_data_dict(db, tx_kv_nodes)

            return self._persist_block(db, block, genesis_parent_hash)

    @classmethod
    def _persist_block(
        cls, db: DatabaseAPI, block: BlockAPI, genesis_parent_hash: Hash32
    ) -> Tuple[Tuple[Hash32, ...], Tuple[Hash32, ...]]:
        header_chain = (block.header,)
        new_canonical_headers, old_canonical_headers = cls._persist_header_chain(
            db, header_chain, genesis_parent_hash
        )

        for header in new_canonical_headers:
            if header.hash == block.hash:
                # Most of the time this is called to persist a block whose parent is the
                # current head, so we optimize for that and read the tx hashes from the
                # block itself. This is specially important during a fast sync.
                tx_hashes = tuple(tx.hash for tx in block.transactions)
            else:
                tx_hashes = cls._get_block_transaction_hashes(db, header)

            for index, transaction_hash in enumerate(tx_hashes):
                cls._add_transaction_to_canonical_chain(
                    db, transaction_hash, header, index
                )

            # post-shanghai, look for withdrawals
            if hasattr(block, "withdrawals") and block.withdrawals not in (None, ()):
                withdrawal_hashes = tuple(
                    withdrawal.hash for withdrawal in block.withdrawals
                )
                for index, withdrawal_hash in enumerate(withdrawal_hashes):
                    cls._add_withdrawal_to_canonical_chain(
                        db,
                        withdrawal_hash,
                        header,
                        index,
                    )

        if block.uncles:
            uncles_hash = cls._persist_uncles(db, block.uncles)
        else:
            uncles_hash = EMPTY_UNCLE_HASH
        if uncles_hash != block.header.uncles_hash:
            raise ValidationError(
                "Block's uncles_hash (%s) does not match actual uncles' hash (%s)",
                block.header.uncles_hash,
                uncles_hash,
            )
        new_canonical_hashes = tuple(header.hash for header in new_canonical_headers)
        old_canonical_hashes = tuple(header.hash for header in old_canonical_headers)

        cls._update_chain_gaps(db, block)
        return new_canonical_hashes, old_canonical_hashes

    def persist_uncles(self, uncles: Tuple[BlockHeaderAPI]) -> Hash32:
        return self._persist_uncles(self.db, uncles)

    @staticmethod
    def _persist_uncles(db: DatabaseAPI, uncles: Tuple[BlockHeaderAPI, ...]) -> Hash32:
        uncles_hash = keccak(rlp.encode(uncles))
        db.set(
            uncles_hash,
            rlp.encode(uncles, sedes=rlp.sedes.CountableList(HeaderSedes)),
        )
        return cast(Hash32, uncles_hash)

    #
    # Block Data API (Transactions, Receipts, and Withdrawals)
    #
    @staticmethod
    def _get_block_data_from_root_hash(
        db: DatabaseAPI,
        block_root_hash: Hash32,
    ) -> Iterable[Hash32]:
        """
        Returns iterable of the encoded items from a root hash in a block. This can be
        useful for retrieving encoded transactions or withdrawals from the
        transaction_root or withdrawals_root of a block.
        """
        item_db = HexaryTrie(db, root_hash=block_root_hash)
        for item_idx in itertools.count():
            item_key = rlp.encode(item_idx)
            encoded = item_db[item_key]
            if encoded != b"":
                yield encoded
            else:
                break

    #
    # Transaction API
    #
    def add_receipt(
        self, block_header: BlockHeaderAPI, index_key: int, receipt: ReceiptAPI
    ) -> Hash32:
        receipt_db = HexaryTrie(db=self.db, root_hash=block_header.receipt_root)
        receipt_db[index_key] = receipt.encode()
        return receipt_db.root_hash

    def add_transaction(
        self,
        block_header: BlockHeaderAPI,
        index_key: int,
        transaction: SignedTransactionAPI,
    ) -> Hash32:
        transaction_db = HexaryTrie(self.db, root_hash=block_header.transaction_root)
        transaction_db[index_key] = transaction.encode()
        return transaction_db.root_hash

    def get_block_transactions(
        self, header: BlockHeaderAPI, transaction_decoder: Type[TransactionDecoderAPI]
    ) -> Tuple[SignedTransactionAPI, ...]:
        return self._get_block_transactions(
            header.transaction_root, transaction_decoder
        )

    def get_block_transaction_hashes(
        self, block_header: BlockHeaderAPI
    ) -> Tuple[Hash32, ...]:
        """
        Returns an iterable of the transaction hashes from the block specified
        by the given block header.
        """
        return self._get_block_transaction_hashes(self.db, block_header)

    @classmethod
    @to_tuple
    def _get_block_transaction_hashes(
        cls, db: DatabaseAPI, block_header: BlockHeaderAPI
    ) -> Iterable[Hash32]:
        all_encoded_transactions = cls._get_block_data_from_root_hash(
            db,
            block_header.transaction_root,
        )
        for encoded_transaction in all_encoded_transactions:
            yield cast(Hash32, keccak(encoded_transaction))

    @to_tuple
    def get_receipts(
        self, header: BlockHeaderAPI, receipt_decoder: Type[ReceiptDecoderAPI]
    ) -> Iterable[ReceiptAPI]:
        receipt_db = HexaryTrie(db=self.db, root_hash=header.receipt_root)
        for receipt_idx in itertools.count():
            receipt_key = rlp.encode(receipt_idx)
            receipt_data = receipt_db[receipt_key]
            if receipt_data != b"":
                yield receipt_decoder.decode(receipt_data)
            else:
                break

    def get_transaction_by_index(
        self,
        block_number: BlockNumber,
        transaction_index: int,
        transaction_decoder: Type[TransactionDecoderAPI],
    ) -> SignedTransactionAPI:
        try:
            block_header = self.get_canonical_block_header_by_number(block_number)
        except HeaderNotFound:
            raise TransactionNotFound(
                f"Block {block_number} is not in the canonical chain"
            )
        transaction_db = HexaryTrie(self.db, root_hash=block_header.transaction_root)
        encoded_index = rlp.encode(transaction_index)
        encoded_transaction = transaction_db[encoded_index]
        if encoded_transaction != b"":
            return transaction_decoder.decode(encoded_transaction)
        else:
            raise TransactionNotFound(
                f"No transaction is at index {transaction_index} "
                f"of block {block_number}"
            )

    def get_transaction_index(
        self, transaction_hash: Hash32
    ) -> Tuple[BlockNumber, int]:
        key = SchemaV1.make_transaction_hash_to_block_lookup_key(transaction_hash)
        try:
            encoded_key = self.db[key]
        except KeyError:
            raise TransactionNotFound(
                f"Transaction {encode_hex(transaction_hash)} "
                "not found in canonical chain"
            )

        transaction_key = rlp.decode(encoded_key, sedes=BlockDataKey)
        return (transaction_key.block_number, transaction_key.index)

    def get_receipt_by_index(
        self,
        block_number: BlockNumber,
        receipt_index: int,
        receipt_decoder: Type[ReceiptDecoderAPI],
    ) -> ReceiptAPI:
        try:
            block_header = self.get_canonical_block_header_by_number(block_number)
        except HeaderNotFound:
            raise ReceiptNotFound(f"Block {block_number} is not in the canonical chain")

        receipt_db = HexaryTrie(db=self.db, root_hash=block_header.receipt_root)
        receipt_key = rlp.encode(receipt_index)
        receipt_data = receipt_db[receipt_key]
        if receipt_data != b"":
            return receipt_decoder.decode(receipt_data)
        else:
            raise ReceiptNotFound(
                f"Receipt with index {receipt_index} not found in block"
            )

    @functools.lru_cache(maxsize=32)  # noqa: B019
    @to_tuple
    def _get_block_transactions(
        self, transaction_root: Hash32, transaction_decoder: Type[TransactionDecoderAPI]
    ) -> Iterable[SignedTransactionAPI]:
        """
        Memoizable version of `get_block_transactions`
        """
        for encoded_transaction in self._get_block_data_from_root_hash(
            self.db, transaction_root
        ):
            yield transaction_decoder.decode(encoded_transaction)

    @staticmethod
    def _remove_transaction_from_canonical_chain(
        db: DatabaseAPI, transaction_hash: Hash32
    ) -> None:
        """
        Removes the transaction specified by the given hash from the canonical
        chain.
        """
        db.delete(SchemaV1.make_transaction_hash_to_block_lookup_key(transaction_hash))

    @staticmethod
    def _add_transaction_to_canonical_chain(
        db: DatabaseAPI,
        transaction_hash: Hash32,
        block_header: BlockHeaderAPI,
        index: int,
    ) -> None:
        """
        :param bytes transaction_hash: the hash of the transaction to add the lookup for
        :param block_header: The header of the block with the txn that is in the
            canonical chain
        :param int index: the position of the transaction in the block
        - add lookup from transaction hash to the block number and index that the body
            is stored at
        - remove transaction hash to body lookup in the pending pool
        """
        transaction_key = BlockDataKey(block_header.block_number, index)
        db.set(
            SchemaV1.make_transaction_hash_to_block_lookup_key(transaction_hash),
            rlp.encode(transaction_key),
        )

    #
    # Withdrawals API
    #
    def get_block_withdrawals(
        self,
        header: BlockHeaderAPI,
    ) -> Tuple[WithdrawalAPI, ...]:
        return self._get_block_withdrawals(header.withdrawals_root)

    @functools.lru_cache(maxsize=32)  # noqa: B019
    @to_tuple
    def _get_block_withdrawals(
        self,
        withdrawals_root: Hash32,
    ) -> Iterable[WithdrawalAPI]:
        """
        Memoizable version of `get_block_withdrawals`
        """
        for encoded_withdrawal in self._get_block_data_from_root_hash(
            self.db,
            withdrawals_root,
        ):
            yield rlp.decode(encoded_withdrawal, sedes=Withdrawal)

    @staticmethod
    def _add_withdrawal_to_canonical_chain(
        db: DatabaseAPI,
        withdrawal_hash: Hash32,
        block_header: BlockHeaderAPI,
        index: int,
    ) -> None:
        withdrawal_key = BlockDataKey(block_header.block_number, index)
        db.set(
            SchemaV1.make_withdrawal_hash_to_block_lookup_key(withdrawal_hash),
            rlp.encode(withdrawal_key),
        )

    #
    # Raw Database API
    #
    def exists(self, key: bytes) -> bool:
        return self.db.exists(key)

    def get(self, key: bytes) -> bytes:
        return self.db[key]

    def persist_trie_data_dict(self, trie_data_dict: Dict[Hash32, bytes]) -> None:
        with self.db.atomic_batch() as db:
            self._persist_trie_data_dict(db, trie_data_dict)

    @classmethod
    def _persist_trie_data_dict(
        cls, db: DatabaseAPI, trie_data_dict: Dict[Hash32, bytes]
    ) -> None:
        for key, value in trie_data_dict.items():
            db[key] = value



from typing import (
    Type,
)

from eth.abc import BlockAPI, BlockHeaderAPI, ConsensusAPI
from eth.consensus.pos import PosConsensus
from eth.constants import (
    POST_MERGE_DIFFICULTY,
    POST_MERGE_NONCE,
    POST_MERGE_OMMERS_HASH,
)
from eth.rlp.blocks import BaseBlock
from eth.vm.forks.gray_glacier import GrayGlacierVM
from eth.vm.state import BaseState
from eth_utils import (
    ValidationError,
)

from .blocks import ParisBlock
from .headers import (
    configure_paris_header,
    create_paris_header_from_parent,
)
from .state import ParisState


class ParisVM(GrayGlacierVM):
    # fork name
    fork = "paris"

    # classes
    block_class: Type[BaseBlock] = ParisBlock
    _state_class: Type[BaseState] = ParisState
    consensus_class: Type[ConsensusAPI] = PosConsensus

    # Methods
    create_header_from_parent = staticmethod(  # type: ignore
        create_paris_header_from_parent()
    )
    configure_header = configure_paris_header

    @staticmethod
    def get_block_reward() -> int:
        return 0

    @classmethod
    def get_nephew_reward(cls) -> int:
        return 0

    def _assign_block_rewards(self, block: BlockAPI) -> None:
        # No block reward or uncles / uncle rewards in PoS
        pass

    @classmethod
    def validate_header(
        cls, header: BlockHeaderAPI, parent_header: BlockHeaderAPI
    ) -> None:
        super().validate_header(header, parent_header)

        difficulty, nonce, uncles_hash = (
            header.difficulty,
            header.nonce,
            header.uncles_hash,
        )

        if difficulty != POST_MERGE_DIFFICULTY:
            raise ValidationError(
                f"Difficulty must be {POST_MERGE_DIFFICULTY}, got {difficulty}."
            )
        if nonce != POST_MERGE_NONCE:
            raise ValidationError(
                f"Nonce must be {POST_MERGE_NONCE !r}, got {nonce !r}."
            )
        if uncles_hash != POST_MERGE_OMMERS_HASH:
            raise ValidationError(
                f"Uncles hash must be {POST_MERGE_OMMERS_HASH !r}, "
                f"got {uncles_hash !r}."
            )



from typing import (
    Any,
    TypeVar,
    Union,
)

from eth_utils.toolz import (
    merge,
)

from eth.abc import (
    SignedTransactionAPI,
    UnsignedTransactionAPI,
)
from eth.constants import (
    DEFAULT_SPOOF_R,
    DEFAULT_SPOOF_S,
    DEFAULT_SPOOF_Y_PARITY,
)

SPOOF_ATTRIBUTES_DEFAULTS = {
    "y_parity": DEFAULT_SPOOF_Y_PARITY,
    "r": DEFAULT_SPOOF_R,
    "s": DEFAULT_SPOOF_S,
}

T = TypeVar("T", bound="SpoofAttributes")


class SpoofAttributes:
    def __init__(
        self,
        spoof_target: Union[SignedTransactionAPI, UnsignedTransactionAPI],
        **overrides: Any
    ) -> None:
        self.spoof_target = spoof_target
        self.overrides = overrides

        if "from_" in overrides:
            if hasattr(spoof_target, "sender"):
                raise TypeError(
                    "A from_ parameter can only be supplied when the spoof target "
                    "does not have a sender attribute.  SpoofTransaction will not "
                    "attempt to override the sender of a signed transaction."
                )

            overrides["sender"] = overrides["from_"]
            overrides["get_sender"] = lambda: overrides["from_"]
            for attr, value in SPOOF_ATTRIBUTES_DEFAULTS.items():
                if not hasattr(spoof_target, attr):
                    overrides[attr] = value

    def __getattr__(self, attr: str) -> Any:
        if attr in self.overrides:
            return self.overrides[attr]
        else:
            return getattr(self.spoof_target, attr)

    def copy(self: T, **kwargs: Any) -> T:
        new_target = self.spoof_target.copy(**kwargs)
        new_overrides = merge(self.overrides, kwargs)
        return type(self)(new_target, **new_overrides)



import functools
from typing import (
    Iterable,
    Sequence,
    Tuple,
    cast,
)

from eth_typing import (
    BlockNumber,
    Hash32,
    Address
)
from eth_utils import (
    ValidationError,
    encode_hex,
    to_tuple,
)
from eth_utils.toolz import (
    concat,
    first,
    sliding_window,
)
import rlp
from web3 import Web3

from eth.abc import (
    AtomicDatabaseAPI,
    BlockHeaderAPI,
    DatabaseAPI,
    HeaderDatabaseAPI,
)
from eth.constants import (
    GENESIS_PARENT_HASH,
)
from eth.db.chain_gaps import (
    GAP_WRITES,
    GENESIS_CHAIN_GAPS,
    GapChange,
    GapInfo,
    fill_gap,
    reopen_gap,
)
from eth.db.schema import (
    SchemaV1,
)
from eth.exceptions import (
    CanonicalHeadNotFound,
    CheckpointsMustBeCanonical,
    HeaderNotFound,
    ParentNotFound,
)
from eth.rlp.headers import BlockHeader
from eth.rlp.sedes import (
    chain_gaps,
)
from eth.typing import (
    ChainGaps,
)
from eth.validation import (
    validate_block_number,
    validate_word,
)
from eth.vm.forks.arrow_glacier.blocks import ArrowGlacierBlockHeader
from eth.vm.forks.gray_glacier.blocks import GrayGlacierBlockHeader
from eth.vm.forks.london.blocks import LondonBlockHeader
from eth.vm.forks.paris.blocks import ParisBlockHeader
from eth.vm.forks.shanghai.blocks import ShanghaiBlockHeader
from eth.vm.header import (
    HeaderSedes,
)
from web3.types import (
    BlockIdentifier,
)

class HeaderDB(HeaderDatabaseAPI):
    w3: Web3

    def __init__(self, w3: Web3) -> None:
        assert w3 is not None
        self.w3 = w3

    def get_header_chain_gaps(self) -> ChainGaps:
        return self._get_header_chain_gaps(self.db)

    @classmethod
    def _get_header_chain_gaps(cls, db: DatabaseAPI) -> ChainGaps:
        try:
            encoded_gaps = db[SchemaV1.make_header_chain_gaps_lookup_key()]
        except KeyError:
            return GENESIS_CHAIN_GAPS
        else:
            return rlp.decode(encoded_gaps, sedes=chain_gaps)

    @classmethod
    def _update_header_chain_gaps(
        cls,
        db: DatabaseAPI,
        persisted_header: BlockHeaderAPI,
        base_gaps: ChainGaps = None,
    ) -> GapInfo:
        # If we make many updates in a row, we can avoid reloading the integrity info by
        # continuously caching it and providing it as a parameter to this API
        if base_gaps is None:
            base_gaps = cls._get_header_chain_gaps(db)

        gap_change, gaps = fill_gap(persisted_header.block_number, base_gaps)

        if gap_change is not GapChange.NoChange:
            db.set(
                SchemaV1.make_header_chain_gaps_lookup_key(),
                rlp.encode(gaps, sedes=chain_gaps),
            )

        return gap_change, gaps

    #
    # Canonical Chain API
    #
    def get_canonical_block_hash(self, block_number: BlockNumber) -> Hash32:
        return self._get_canonical_block_hash(self.db, block_number)

    @staticmethod
    def _get_canonical_block_hash(db: DatabaseAPI, block_number: BlockNumber) -> Hash32:
        validate_block_number(block_number)
        number_to_hash_key = SchemaV1.make_block_number_to_hash_lookup_key(block_number)

        try:
            encoded_key = db[number_to_hash_key]
        except KeyError:
            raise HeaderNotFound(
                f"No canonical header for block number #{block_number}"
            )
        else:
            return rlp.decode(encoded_key, sedes=rlp.sedes.binary)

    def get_canonical_block_header_by_number(
        self, block_number: BlockNumber
    ) -> BlockHeaderAPI:
        return self._get_canonical_block_header_by_number(self.db, block_number)

    @classmethod
    def _get_canonical_block_header_by_number(
        cls, db: DatabaseAPI, block_number: BlockNumber
    ) -> BlockHeaderAPI:
        validate_block_number(block_number)
        canonical_block_hash = cls._get_canonical_block_hash(db, block_number)
        return cls._get_block_header_by_hash(db, canonical_block_hash)

    def get_canonical_head(self) -> BlockHeaderAPI:
        return self._get_canonical_head(self.db)

    @classmethod
    def _get_canonical_head(cls, db: DatabaseAPI) -> BlockHeaderAPI:
        canonical_head_hash = cls._get_canonical_head_hash(db)
        return cls._get_block_header_by_hash(db, canonical_head_hash)

    @classmethod
    def _get_canonical_head_hash(cls, db: DatabaseAPI) -> Hash32:
        try:
            return Hash32(db[SchemaV1.make_canonical_head_hash_lookup_key()])
        except KeyError:
            raise CanonicalHeadNotFound("No canonical head set for this chain")

    #
    # Header API
    #
    def get_block_header_by_hash(self, block_hash: Hash32) -> BlockHeaderAPI:
        import pdb; pdb.set_trace()
        assert self.w3 is not None
        return build_block_header(self.w3, block_hash)
        return self._get_block_header_by_hash(self.db, block_hash)

    # @staticmethod
    # def _get_block_header_by_hash(
    #     db: DatabaseAPI, block_hash: Hash32
    # ) -> BlockHeaderAPI:
    #     """
    #     Returns the requested block header as specified by block hash.

    #     Raises BlockNotFound if it is not present in the db.
    #     """
    #     validate_word(block_hash, title="Block Hash")
    #     try:
    #         header_rlp = db[block_hash]
    #     except KeyError:
    #         raise HeaderNotFound(f"No header with hash {encode_hex(block_hash)} found")
    #     return _decode_block_header(header_rlp)

    def get_score(self, block_hash: Hash32) -> int:
        return self._get_score(self.db, block_hash)

    @staticmethod
    def _get_score(db: DatabaseAPI, block_hash: Hash32) -> int:
        try:
            encoded_score = db[SchemaV1.make_block_hash_to_score_lookup_key(block_hash)]
        except KeyError:
            raise HeaderNotFound(f"No header with hash {encode_hex(block_hash)} found")
        return rlp.decode(encoded_score, sedes=rlp.sedes.big_endian_int)

    def header_exists(self, block_hash: Hash32) -> bool:
        return self._header_exists(self.db, block_hash)

    @staticmethod
    def _header_exists(db: DatabaseAPI, block_hash: Hash32) -> bool:
        validate_word(block_hash, title="Block Hash")
        return block_hash in db

    def persist_header(
        self, header: BlockHeaderAPI
    ) -> Tuple[Tuple[BlockHeaderAPI, ...], Tuple[BlockHeaderAPI, ...]]:
        return self.persist_header_chain((header,))

    def persist_header_chain(
        self,
        headers: Iterable[BlockHeaderAPI],
        genesis_parent_hash: Hash32 = GENESIS_PARENT_HASH,
    ) -> Tuple[Tuple[BlockHeaderAPI, ...], Tuple[BlockHeaderAPI, ...]]:
        with self.db.atomic_batch() as db:
            return self._persist_header_chain(db, headers, genesis_parent_hash)

    def persist_checkpoint_header(self, header: BlockHeaderAPI, score: int) -> None:
        with self.db.atomic_batch() as db:
            return self._persist_checkpoint_header(db, header, score)

    @classmethod
    def _set_hash_scores_to_db(
        cls, db: DatabaseAPI, header: BlockHeaderAPI, score: int
    ) -> int:
        difficulty = header.difficulty
        new_score = (
            # In PoS, difficulty = 0 and score values do not need to work the same way
            score + difficulty
            if difficulty != 0
            else score + header.block_number
        )

        db.set(
            SchemaV1.make_block_hash_to_score_lookup_key(header.hash),
            rlp.encode(new_score, sedes=rlp.sedes.big_endian_int),
        )

        return new_score

    @classmethod
    def _persist_checkpoint_header(
        cls, db: DatabaseAPI, header: BlockHeaderAPI, score: int
    ) -> None:
        db.set(
            header.hash,
            rlp.encode(header),
        )

        # Add new checkpoint header
        previous_checkpoints = cls._get_checkpoints(db)
        new_checkpoints = previous_checkpoints + (header.hash,)
        db.set(
            SchemaV1.make_checkpoint_headers_key(),
            b"".join(new_checkpoints),
        )

        difficulty = header.difficulty
        previous_score = (
            # In PoS, difficulty = 0 and score values do not need to work the same way
            score - difficulty
            if difficulty != 0
            else score - header.block_number
        )
        cls._set_hash_scores_to_db(db, header, previous_score)
        cls._set_as_canonical_chain_head(db, header, GENESIS_PARENT_HASH)
        _, gaps = cls._update_header_chain_gaps(db, header)

        # check if the parent block number exists, and is not a match
        # for checkpoint.parent_hash
        parent_block_num = BlockNumber(header.block_number - 1)
        try:
            parent_hash = cls._get_canonical_block_hash(db, parent_block_num)
        except HeaderNotFound:
            # no parent to check
            pass
        else:
            # User is asserting that the checkpoint must be canonical, so if the parent
            # doesn't match, then the parent must not be canonical,
            # and should be de-canonicalized.
            if parent_hash != header.parent_hash:
                # does the correct header exist in the database?
                try:
                    true_parent = cls._get_block_header_by_hash(db, header.parent_hash)
                except HeaderNotFound:
                    # True parent unavailable, just delete the now non-canonical one
                    cls._decanonicalize_single(db, parent_block_num, gaps)
                else:
                    # True parent should have already been canonicalized during
                    #   _set_as_canonical_chain_head()
                    raise ValidationError(
                        f"Why was a non-matching parent header {parent_hash!r} left as "
                        "canonical after _set_as_canonical_chain_head() and "
                        f"{true_parent} is available?"
                    )

        cls._decanonicalize_descendant_orphans(db, header, new_checkpoints)

    @classmethod
    def _decanonicalize_descendant_orphans(
        cls, db: DatabaseAPI, header: BlockHeaderAPI, checkpoints: Tuple[Hash32, ...]
    ) -> None:
        # Determine if any children need to be de-canonicalized because they are not
        # children of the new chain head
        new_gaps = starting_gaps = cls._get_header_chain_gaps(db)

        child_number = BlockNumber(header.block_number + 1)
        try:
            child = cls._get_canonical_block_header_by_number(db, child_number)
        except HeaderNotFound:
            # There is no canonical block here
            next_invalid_child = None
        else:
            if child.parent_hash != header.hash:
                if child.hash in checkpoints:
                    raise CheckpointsMustBeCanonical(
                        f"Trying to decanonicalize {child} while making "
                        f"{header} the chain tip"
                    )
                else:
                    next_invalid_child = child
            else:
                next_invalid_child = None

        while next_invalid_child:
            # decanonicalize, and add gap for tracking
            db.delete(SchemaV1.make_block_number_to_hash_lookup_key(child_number))
            new_gaps = reopen_gap(child_number, new_gaps)

            # find next child
            child_number = BlockNumber(child_number + 1)
            try:
                # All contiguous children must now be made invalid
                next_invalid_child = cls._get_canonical_block_header_by_number(
                    db, child_number
                )
            except HeaderNotFound:
                # Found the end of this streak of canonical blocks
                break
            else:
                if next_invalid_child.hash in checkpoints:
                    raise CheckpointsMustBeCanonical(
                        f"Trying to decanonicalize {next_invalid_child} while making "
                        f"{header} the chain tip"
                    )

        if new_gaps != starting_gaps:
            db.set(
                SchemaV1.make_header_chain_gaps_lookup_key(),
                rlp.encode(new_gaps, sedes=chain_gaps),
            )

    @classmethod
    def _decanonicalize_single(
        cls, db: DatabaseAPI, block_num: BlockNumber, base_gaps: ChainGaps
    ) -> ChainGaps:
        """
        A single block number was found to no longer be canonical. At doc-time,
        this only happens because it does not link up with a checkpoint header.
        So de-canonicalize this block number and insert a gap in the tracked
        chain gaps.
        """

        db.delete(SchemaV1.make_block_number_to_hash_lookup_key(block_num))

        new_gaps = reopen_gap(block_num, base_gaps)
        if new_gaps != base_gaps:
            db.set(
                SchemaV1.make_header_chain_gaps_lookup_key(),
                rlp.encode(new_gaps, sedes=chain_gaps),
            )
        return new_gaps

    @classmethod
    def _persist_header_chain(
        cls,
        db: DatabaseAPI,
        headers: Iterable[BlockHeaderAPI],
        genesis_parent_hash: Hash32,
    ) -> Tuple[Tuple[BlockHeaderAPI, ...], Tuple[BlockHeaderAPI, ...]]:
        headers_iterator = iter(headers)

        try:
            first_header = first(headers_iterator)
        except StopIteration:
            return (), ()

        is_genesis = first_header.parent_hash == genesis_parent_hash
        if not is_genesis and not cls._header_exists(db, first_header.parent_hash):
            raise ParentNotFound(
                f"Cannot persist block header ({encode_hex(first_header.hash)}) "
                f"with unknown parent ({encode_hex(first_header.parent_hash)})"
            )

        if is_genesis:
            score = 0
        else:
            score = cls._get_score(db, first_header.parent_hash)

        curr_chain_head = first_header
        db.set(
            curr_chain_head.hash,
            rlp.encode(curr_chain_head),
        )
        score = cls._set_hash_scores_to_db(db, curr_chain_head, score)

        base_gaps = cls._get_header_chain_gaps(db)
        gap_info = cls._update_header_chain_gaps(db, curr_chain_head, base_gaps)
        gaps = cls._handle_gap_change(
            db, gap_info, curr_chain_head, genesis_parent_hash
        )

        orig_headers_seq = concat([(first_header,), headers_iterator])
        for parent, child in sliding_window(2, orig_headers_seq):
            if parent.hash != child.parent_hash:
                raise ValidationError(
                    f"Non-contiguous chain. Expected {encode_hex(child.hash)} "
                    f"to have {encode_hex(parent.hash)} as parent "
                    f"but was {encode_hex(child.parent_hash)}"
                )

            curr_chain_head = child
            db.set(
                curr_chain_head.hash,
                rlp.encode(curr_chain_head),
            )

            score = cls._set_hash_scores_to_db(db, curr_chain_head, score)
            gap_info = cls._update_header_chain_gaps(db, curr_chain_head, gaps)
            gaps = cls._handle_gap_change(
                db, gap_info, curr_chain_head, genesis_parent_hash
            )
        try:
            previous_canonical_head = cls._get_canonical_head_hash(db)
            head_score = cls._get_score(db, previous_canonical_head)
        except CanonicalHeadNotFound:
            return cls._set_as_canonical_chain_head(
                db, curr_chain_head, genesis_parent_hash
            )

        if score > head_score:
            return cls._set_as_canonical_chain_head(
                db, curr_chain_head, genesis_parent_hash
            )

        return (), ()

    @classmethod
    def _handle_gap_change(
        cls,
        db: DatabaseAPI,
        gap_info: GapInfo,
        header: BlockHeaderAPI,
        genesis_parent_hash: Hash32,
    ) -> ChainGaps:
        gap_change, gaps = gap_info
        if gap_change not in GAP_WRITES:
            return gaps

        # Check if this change will link up the chain to the right
        if gap_change in (GapChange.GapFill, GapChange.GapRightShrink):
            next_child_number = BlockNumber(header.block_number + 1)
            expected_child = cls._get_canonical_block_header_by_number(
                db, next_child_number
            )
            if header.hash != expected_child.parent_hash:
                # Must not join a canonical chain that is not linked from parent to
                # child. If the child is a checkpoint, reject this fill as an uncle.
                checkpoints = cls._get_checkpoints(db)
                if expected_child.hash in checkpoints:
                    raise CheckpointsMustBeCanonical(
                        f"Cannot make {header} canonical, because it is not the parent "
                        f"of declared checkpoint: {expected_child}"
                    )
                else:
                    # If the child is *not* a checkpoint,
                    # then re-open a gap in the chain
                    gaps = cls._decanonicalize_single(
                        db, expected_child.block_number, gaps
                    )

        # We implicitly assert that persisted headers are canonical here.
        # This assertion is made when persisting headers that are known to be part of a
        # gap in the canonical chain.
        # What if this assertion is later found to be false? At gap fill time, we can
        # detect if the chains don't link (and raise a ValidationError). Also, when a
        # true canonical header is added eventually, we need to canonicalize all the
        # true headers.
        cls._canonicalize_header(db, header, genesis_parent_hash)
        return gaps

    @classmethod
    def _canonicalize_header(
        cls,
        db: DatabaseAPI,
        header: BlockHeaderAPI,
        genesis_parent_hash: Hash32,
    ) -> Tuple[Tuple[BlockHeaderAPI, ...], Tuple[BlockHeaderAPI, ...]]:
        """
        Force this header to be canonical,
        and adjust its ancestors/descendants as necessary

        :raises CheckpointsMustBeCanonical: if trying to set a head that would
            de-canonicalize a checkpoint
        """
        new_canonical_headers = cast(
            Tuple[BlockHeaderAPI, ...],
            tuple(reversed(cls._find_new_ancestors(db, header, genesis_parent_hash))),
        )
        old_canonical_headers = cls._find_headers_to_decanonicalize(
            db,
            [h.block_number for h in new_canonical_headers],
        )

        # Reject if this would make a checkpoint non-canonical
        checkpoints = cls._get_checkpoints(db)
        attempted_checkpoint_overrides = {
            old for old in old_canonical_headers if old.hash in checkpoints
        }
        if len(attempted_checkpoint_overrides):
            raise CheckpointsMustBeCanonical(
                "Tried to switch chain away from checkpoint(s) "
                f"{attempted_checkpoint_overrides!r} by inserting new canonical "
                f"headers {new_canonical_headers}"
            )

        for ancestor in new_canonical_headers:
            cls._add_block_number_to_hash_lookup(db, ancestor)

        if len(new_canonical_headers):
            cls._decanonicalize_descendant_orphans(
                db, new_canonical_headers[-1], checkpoints
            )

        return new_canonical_headers, old_canonical_headers

    @classmethod
    def _set_as_canonical_chain_head(
        cls,
        db: DatabaseAPI,
        header: BlockHeaderAPI,
        genesis_parent_hash: Hash32,
    ) -> Tuple[Tuple[BlockHeaderAPI, ...], Tuple[BlockHeaderAPI, ...]]:
        """
        Sets the canonical chain HEAD to the block header as specified by the
        given block hash.

        :return: a tuple of the headers that are newly in the canonical chain, and the
            headers that are no longer in the canonical chain
        :raises CheckpointsMustBeCanonical: if trying to set a head that would
            de-canonicalize a checkpoint
        """
        try:
            current_canonical_head = cls._get_canonical_head_hash(db)
        except CanonicalHeadNotFound:
            current_canonical_head = None

        new_canonical_headers: Tuple[BlockHeaderAPI, ...]
        old_canonical_headers: Tuple[BlockHeaderAPI, ...]

        if current_canonical_head and header.parent_hash == current_canonical_head:
            # the calls to _find_new_ancestors and _find_headers_to_decanonicalize are
            # relatively expensive, it's better to skip them in this case, where we're
            # extending the canonical chain by a header
            new_canonical_headers = (header,)
            old_canonical_headers = ()
            cls._add_block_number_to_hash_lookup(db, header)
        else:
            (
                new_canonical_headers,
                old_canonical_headers,
            ) = cls._canonicalize_header(db, header, genesis_parent_hash)

        db.set(SchemaV1.make_canonical_head_hash_lookup_key(), header.hash)

        return new_canonical_headers, old_canonical_headers

    @classmethod
    def _get_checkpoints(cls, db: DatabaseAPI) -> Tuple[Hash32, ...]:
        concatenated_checkpoints = db.get(SchemaV1.make_checkpoint_headers_key())
        if concatenated_checkpoints is None:
            return ()
        else:
            return tuple(
                Hash32(concatenated_checkpoints[index : index + 32])
                for index in range(0, len(concatenated_checkpoints), 32)
            )

    @classmethod
    @to_tuple
    def _find_headers_to_decanonicalize(
        cls, db: DatabaseAPI, numbers_to_decanonicalize: Sequence[BlockNumber]
    ) -> Iterable[BlockHeaderAPI]:
        for block_number in numbers_to_decanonicalize:
            try:
                old_canonical_hash = cls._get_canonical_block_hash(db, block_number)
            except HeaderNotFound:
                # no old_canonical block, but due to checkpointing, more may be possible
                continue
            else:
                yield cls._get_block_header_by_hash(db, old_canonical_hash)

    @classmethod
    @to_tuple
    def _find_new_ancestors(
        cls, db: DatabaseAPI, header: BlockHeaderAPI, genesis_parent_hash: Hash32
    ) -> Iterable[BlockHeaderAPI]:
        """
        Returns the chain leading up from the given header until (but not including)
        the first ancestor it has in common with our canonical chain.

        If D is the canonical head in the following chain, and F is the new header,
        then this function returns (F, E).

        A - B - C - D
               \
                E - F
        """
        h = header
        while True:
            try:
                orig = cls._get_canonical_block_header_by_number(db, h.block_number)
            except HeaderNotFound:
                # This just means the block is not on the canonical chain.
                pass
            else:
                if orig.hash == h.hash:
                    # Found the common ancestor, stop.
                    break

            # Found a new ancestor
            yield h

            if h.parent_hash == genesis_parent_hash:
                break
            else:
                try:
                    h = cls._get_block_header_by_hash(db, h.parent_hash)
                except HeaderNotFound:
                    # We must have hit a checkpoint parent, return early
                    break

    @staticmethod
    def _add_block_number_to_hash_lookup(
        db: DatabaseAPI, header: BlockHeaderAPI
    ) -> None:
        """
        Sets a record in the database to allow looking up this header by its
        block number.
        """
        block_number_to_hash_key = SchemaV1.make_block_number_to_hash_lookup_key(
            header.block_number
        )
        db.set(
            block_number_to_hash_key,
            rlp.encode(header.hash, sedes=rlp.sedes.binary),
        )


def build_block_header(w3: Web3, block_identifier: BlockIdentifier) -> BlockHeaderAPI:
    """
    Load a block header from geth in the format pyevm likes (not json)
    """
    assert w3 is not None

    # stupid hack to avoid circular imports
    from eth.chains.mainnet.constants import (
        ARROW_GLACIER_MAINNET_BLOCK,
        GRAY_GLACIER_MAINNET_BLOCK,
        PARIS_MAINNET_BLOCK,
        PETERSBURG_MAINNET_BLOCK,
        LONDON_MAINNET_BLOCK,
    )


    # stupid, stupid thing -- there's no way to parse a heaer (in JSON form)
    # directly into a BlockHeader object, so we have to do this.
    # recall that different forks have different header types, so we need to
    # get the right one

    b = w3.eth.get_block(block_identifier)

    block_number = b['number']

    base_args = {
        'difficulty': b['difficulty'],
        'extra_data': b['extraData'],
        'gas_limit': b['gasLimit'],
        'block_number': b['number'],
        'gas_used': b['gasUsed'],
        'timestamp': b['timestamp'],
        'parent_hash': b['parentHash'],
        'uncles_hash': b['sha3Uncles'],
        'bloom': b['logsBloom'],
        'transaction_root': b['transactionsRoot'],
        'receipt_root': b['receiptsRoot'],
        'coinbase': Address(bytes.fromhex(b['miner'][2:])),
        'nonce': b['nonce'],
        'state_root': b['stateRoot'],
    }

    # For Sepolia, let's just use the latest block header for now
    if w3.eth.chain_id == 11155111:
        if block_number >= 2990908:
            return ShanghaiBlockHeader(
                **{
                    **base_args,
                    'base_fee_per_gas': b['baseFeePerGas'],
                    'mix_hash': b['mixHash'],
                }
            )
        else:
            return LondonBlockHeader(
            **{
                **base_args,
                'base_fee_per_gas': b['baseFeePerGas'],
            }
        )

    assert block_number >= PETERSBURG_MAINNET_BLOCK

    if block_number < LONDON_MAINNET_BLOCK:
        ret = BlockHeader(
            **base_args,
        )
    elif block_number < ARROW_GLACIER_MAINNET_BLOCK:
        ret = LondonBlockHeader(
            **{
                **base_args,
                'base_fee_per_gas': b['baseFeePerGas'],
            }
        )
    elif block_number < GRAY_GLACIER_MAINNET_BLOCK:
        ret = ArrowGlacierBlockHeader(
            **{
                **base_args,
                'base_fee_per_gas': b['baseFeePerGas'],
            }
        )
    elif block_number < PARIS_MAINNET_BLOCK:
        ret = GrayGlacierBlockHeader(
            **{
                **base_args,
                'base_fee_per_gas': b['baseFeePerGas'],
                'mix_hash': b['mixHash'],
            }
        )
    elif block_number < 17034870:
        ret = ParisBlockHeader(
            **{
                **base_args,
                'base_fee_per_gas': b['baseFeePerGas'],
                'mix_hash': b['mixHash'],
            }
        )
    else:
        ret = ShanghaiBlockHeader(
            **{
                **base_args,
                'base_fee_per_gas': b['baseFeePerGas'],
                'mix_hash': b['mixHash'],
            }
        )

    return ret

    # if hook is None:
    #     return TargetVM

    # TargetStateClass = TargetVM.get_state_class()


    # ret = BlockHeader(
    #     difficulty = b['difficulty'],
    #     extra_data = b['extraData'],
    #     gas_limit = b['gasLimit'],
    #     block_number = b['number'],
    #     gas_used = b['gasUsed'],
    #     timestamp = b['timestamp'],
    #     parent_hash = b['parentHash'],
    #     uncles_hash = b['sha3Uncles'],
    #     bloom = b['logsBloom'],
    #     transaction_root = b['transactionsRoot'],
    #     receipt_root = b['receiptsRoot'],
    #     coinbase = b['miner'],
    #     nonce = b['nonce'],
    #     mix_hash = b['mixHash'],
    #     state_root = b['stateRoot'],
    # )
    
    return ret



import hashlib
from typing import (
    Any,
    Callable,
    Iterable,
)

from eth_utils.toolz import (
    curry,
    identity,
)

from .loading import (
    find_fixtures,
)


#
# Pytest fixture generation
#
def idfn(fixture_params: Iterable[Any]) -> str:
    """
    Function for pytest to produce uniform names for fixtures.
    """
    try:
        return ":".join(str(item) for item in fixture_params)
    except TypeError:
        # In case params are not iterable for some reason...
        return str(fixture_params)


def get_fixtures_file_hash(all_fixture_paths: Iterable[str]) -> str:
    """
    Returns the MD5 hash of the fixture files.  Used for cache busting.
    """
    hasher = hashlib.md5()
    for fixture_path in sorted(all_fixture_paths):
        with open(fixture_path, "rb") as fixture_file:
            hasher.update(fixture_file.read())
    return hasher.hexdigest()


@curry
def generate_fixture_tests(
    metafunc: Any,
    base_fixture_path: str,
    filter_fn: Callable[..., Any] = identity,
    preprocess_fn: Callable[..., Any] = identity,
) -> None:
    """
    Helper function for use with `pytest_generate_tests` which will use the
    pytest caching facilities to reduce the load time for fixture tests.

    - `metafunc` is the parameter from `pytest_generate_tests`
    - `base_fixture_path` is the base path that fixture files will be collected from.
    - `filter_fn` handles ignoring or marking the various fixtures.
       See `filter_fixtures`.
    - `preprocess_fn` handles any preprocessing that should be done on the raw
       fixtures (such as expanding the statetest fixtures to be multiple tests for
       each fork.
    """
    if "fixture_data" in metafunc.fixturenames:
        all_fixtures = find_fixtures(base_fixture_path)

        if not all_fixtures:
            raise AssertionError(
                f"Suspiciously found zero fixtures: {base_fixture_path}"
            )

        filtered_fixtures = filter_fn(preprocess_fn(all_fixtures))

        metafunc.parametrize("fixture_data", filtered_fixtures, ids=idfn)



from collections import (
    namedtuple,
)
from functools import (
    partial,
    wraps,
)
from typing import (
    Any,
    Callable,
    Dict,
)

from eth_utils import (
    apply_formatters_to_dict,
    decode_hex,
    to_canonical_address,
)
from eth_utils.toolz import (
    assoc,
    assoc_in,
    curry,
    merge,
)

from eth.tools._utils.mappings import (
    deep_merge,
)
from eth.tools._utils.normalization import (
    normalize_environment,
    normalize_execution,
    normalize_networks,
    normalize_state,
    normalize_transaction,
)
from eth.tools._utils.vyper import (
    compile_vyper_lll,
)
from eth.tools.fixtures.helpers import (
    get_test_name,
)
from eth.typing import (
    GeneralState,
    TransactionDict,
)

from ._utils import (
    add_transaction_to_group,
    wrap_in_list,
)

#
# Defaults
#

DEFAULT_MAIN_ENVIRONMENT = {
    "currentCoinbase": to_canonical_address(
        "0x2adc25665018aa1fe0e6bc666dac8fc2697ff9ba"
    ),
    "currentDifficulty": 131072,
    "currentGasLimit": 1000000,
    "currentNumber": 1,
    "currentTimestamp": 1000,
    "previousHash": decode_hex(
        "0x5e20a0453cecd065ea59c37ac63e079ee08998b6045136a8ce6635c7912ec0b6"
    ),
}


DEFAULT_MAIN_TRANSACTION: TransactionDict = {
    "data": b"",
    "gasLimit": 100000,
    "gasPrice": 0,
    "nonce": 0,
    "secretKey": decode_hex(
        "0x45a915e4d060149eb4365960e6a7a45f334393093061116b197e3240065ff2d8"
    ),
    "to": to_canonical_address("0x0f572e5295c57f15886f9b263e2f6d2d6c7b5ec6"),
    "value": 0,
}


def get_default_transaction(networks: Any) -> TransactionDict:
    return DEFAULT_MAIN_TRANSACTION


DEFAULT_EXECUTION = {
    "address": to_canonical_address("0x0f572e5295c57f15886f9b263e2f6d2d6c7b5ec6"),
    "origin": to_canonical_address("0xcd1722f2947def4cf144679da39c4c32bdc35681"),
    "caller": to_canonical_address("0xcd1722f2947def4cf144679da39c4c32bdc35681"),
    "value": 1000000000000000000,
    "data": b"",
    "gasPrice": 1,
    "gas": 100000,
}


Test = namedtuple("Test", ["filler", "fill_kwargs"])
# make `None` default for fill_kwargs
Test.__new__.__defaults__ = (None,)


#
# Filler Generation
#


def setup_filler(
    name: str, environment: Dict[Any, Any] = None
) -> Dict[str, Dict[str, Any]]:
    environment = normalize_environment(environment or {})
    return {
        name: {
            "env": environment,
            "pre": {},
        }
    }


def setup_main_filler(
    name: str, environment: Dict[Any, Any] = None
) -> Dict[str, Dict[str, Any]]:
    """
    Kick off the filler generation process by creating the general filler scaffold with
    a test name and general information about the testing environment.

    For tests for the main chain, the `environment` parameter is expected to be a
    dictionary with some or all of the following keys:

    +------------------------+---------------------------------+
    | key                    | description                     |
    +========================+=================================+
    | ``"currentCoinbase"``  | the coinbase address            |
    +------------------------+---------------------------------+
    | ``"currentNumber"``    | the block number                |
    +------------------------+---------------------------------+
    | ``"previousHash"``     | the hash of the parent block    |
    +------------------------+---------------------------------+
    | ``"currentDifficulty"``| the block's difficulty          |
    +------------------------+---------------------------------+
    | ``"currentGasLimit"``  | the block's gas limit           |
    +------------------------+---------------------------------+
    | ``"currentTimestamp"`` | the timestamp of the block      |
    +------------------------+---------------------------------+
    """
    return setup_filler(name, merge(DEFAULT_MAIN_ENVIRONMENT, environment or {}))


def pre_state(*raw_state: GeneralState, filler: Dict[str, Any]) -> None:
    """
    Specify the state prior to the test execution. Multiple invocations don't override
    the state but extend it instead.

    In general, the elements of `state_definitions` are nested dictionaries of the
    following form:

    .. code-block:: python

        {
            address: {
                "nonce": <account nonce>,
                "balance": <account balance>,
                "code": <account code>,
                "storage": {
                    <storage slot>: <storage value>
                }
            }
        }

    To avoid unnecessary nesting especially if only few fields per account are
    specified, the following and similar formats are possible as well:

    .. code-block:: python

        (address, "balance", <account balance>)
        (address, "storage", <storage slot>, <storage value>)
        (address, "storage", {<storage slot>: <storage value>})
        (address, {"balance", <account balance>})
    """

    @wraps(pre_state)
    def _pre_state(filler: Dict[str, Any]) -> Dict[str, Any]:
        test_name = get_test_name(filler)

        old_pre_state = filler[test_name].get("pre_state", {})
        pre_state = normalize_state(raw_state)
        defaults = {
            address: {
                "balance": 0,
                "nonce": 0,
                "code": b"",
                "storage": {},
            }
            for address in pre_state
        }
        new_pre_state = deep_merge(defaults, old_pre_state, pre_state)

        return assoc_in(filler, [test_name, "pre"], new_pre_state)


def _expect(
    post_state: Dict[str, Any],
    networks: Any,
    transaction: TransactionDict,
    filler: Dict[str, Any],
) -> Dict[str, Any]:
    test_name = get_test_name(filler)
    test = filler[test_name]
    test_update: Dict[str, Dict[Any, Any]] = {test_name: {}}

    pre_state = test.get("pre", {})
    post_state = normalize_state(post_state or {})
    defaults = {
        address: {
            "balance": 0,
            "nonce": 0,
            "code": b"",
            "storage": {},
        }
        for address in post_state
    }
    result = deep_merge(defaults, pre_state, normalize_state(post_state))
    new_expect = {"result": result}

    if transaction is not None:
        transaction = normalize_transaction(
            merge(get_default_transaction(networks), transaction)
        )
        if "transaction" not in test:
            transaction_group = apply_formatters_to_dict(
                {
                    "data": wrap_in_list,
                    "gasLimit": wrap_in_list,
                    "value": wrap_in_list,
                },
                transaction,
            )
            indexes = {
                index_key: 0
                for transaction_key, index_key in [
                    ("gasLimit", "gas"),
                    ("value", "value"),
                    ("data", "data"),
                ]
                if transaction_key in transaction_group
            }
        else:
            transaction_group, indexes = add_transaction_to_group(
                test["transaction"], transaction
            )
        new_expect = assoc(new_expect, "indexes", indexes)
        test_update = assoc_in(
            test_update, [test_name, "transaction"], transaction_group
        )

    if networks is not None:
        networks = normalize_networks(networks)
        new_expect = assoc(new_expect, "networks", networks)

    existing_expects = test.get("expect", [])
    expect = existing_expects + [new_expect]
    test_update = assoc_in(test_update, [test_name, "expect"], expect)

    return deep_merge(filler, test_update)


def expect(
    post_state: Dict[str, Any] = None,
    networks: Any = None,
    transaction: TransactionDict = None,
) -> Callable[..., Dict[str, Any]]:
    """
    Specify the expected result for the test.

    For state tests, multiple expectations can be given, differing in the transaction
    data, gas limit, and value, in the applicable networks, and as a result also in the
    post state. VM tests support only a single expectation with no specified network and
    no transaction.
    (here, its role is played by :func:`~eth.tools.fixtures.fillers.execution`).

    * ``post_state`` is a list of state definition in the same form as expected
      by :func:`~eth.tools.fixtures.fillers.pre_state`. State items that are
      not set explicitly default to their pre state.

    * ``networks`` defines the forks under which the expectation is applicable. It
        should be a sublist of the following identifiers
        (also available in `ALL_FORKS`):

      * ``"Frontier"``
      * ``"Homestead"``
      * ``"EIP150"``
      * ``"EIP158"``
      * ``"Byzantium"``

    * ``transaction`` is a dictionary coming in two variants. For the main shard:

      +----------------+-------------------------------+
      | key            | description                   |
      +================+===============================+
      | ``"data"``     | the transaction data,         |
      +----------------+-------------------------------+
      | ``"gasLimit"`` | the transaction gas limit,    |
      +----------------+-------------------------------+
      | ``"gasPrice"`` | the gas price,                |
      +----------------+-------------------------------+
      | ``"nonce"``    | the transaction nonce,        |
      +----------------+-------------------------------+
      | ``"value"``    | the transaction value         |
      +----------------+-------------------------------+

    In addition, one should specify either the signature itself (via keys ``"v"``,
    ``"r"``, and ``"s"``) or a private key used for signing (via ``"secretKey"``).
    """
    return partial(_expect, post_state, networks, transaction)


@curry
def execution(execution: Dict[str, Any], filler: Dict[str, Any]) -> Dict[str, Any]:
    """
    For VM tests, specify the code that is being run as well as the current state of
    the EVM. State tests don't support this object. The parameter is a dictionary
    specifying some or all of the following keys:

    +--------------------+------------------------------------------------------------+
    |  key               | description                                                |
    +====================+============================================================+
    | ``"address"``      | the address of the account executing the code              |
    +--------------------+------------------------------------------------------------+
    | ``"caller"``       | the caller address                                         |
    +--------------------+------------------------------------------------------------+
    | ``"origin"``       | the origin address (defaulting to the caller address)      |
    +--------------------+------------------------------------------------------------+
    | ``"value"``        | the value of the call                                      |
    +--------------------+------------------------------------------------------------+
    | ``"data"``         | the data passed with the call                              |
    +--------------------+------------------------------------------------------------+
    | ``"gasPrice"``     | the gas price of the call                                  |
    +--------------------+------------------------------------------------------------+
    | ``"gas"``          | the amount of gas allocated for the call                   |
    +--------------------+------------------------------------------------------------+
    | ``"code"``         | the bytecode to execute                                    |
    +--------------------+------------------------------------------------------------+
    | ``"vyperLLLCode"`` | the code in Vyper LLL (compiled to bytecode automatically) |
    +--------------------+------------------------------------------------------------+
    """
    execution = normalize_execution(execution or {})

    # user caller as origin if not explicitly given
    if "caller" in execution and "origin" not in execution:
        execution = assoc(execution, "origin", execution["caller"])

    if "vyperLLLCode" in execution:
        code = compile_vyper_lll(execution["vyperLLLCode"])
        if "code" in execution:
            if code != execution["code"]:
                raise ValueError("Compiled Vyper LLL code does not match")
        execution = assoc(execution, "code", code)

    execution = merge(DEFAULT_EXECUTION, execution)

    test_name = get_test_name(filler)
    return deep_merge(
        filler,
        {
            test_name: {
                "exec": execution,
            }
        },
    )



from typing import (
    Any,
    Dict,
)

from eth_utils.toolz import (
    assoc_in,
    merge,
)

from eth.tools._utils.git import (
    get_version_from_git,
)
from eth.tools.fixtures.helpers import (
    get_test_name,
)

from .formatters import (
    filled_state_test_formatter,
    filled_vm_test_formatter,
)
from .state import (
    fill_state_test,
)
from .vm import (
    fill_vm_test,
)

FILLED_WITH_TEMPLATE = "py-evm-{version}"


#
# Primary test filler
#
def fill_test(
    filler: Dict[str, Any],
    info: Dict[str, Any] = None,
    apply_formatter: bool = True,
    **kwargs: Any
) -> Dict[str, Any]:
    test_name = get_test_name(filler)
    test = filler[test_name]

    if "transaction" in test:
        filled = fill_state_test(filler)
        formatter = filled_state_test_formatter
    elif "exec" in test:
        filled = fill_vm_test(filler, **kwargs)
        formatter = filled_vm_test_formatter
    else:
        raise ValueError("Given filler does not appear to be for VM or state test")

    info = merge(
        {"filledwith": FILLED_WITH_TEMPLATE.format(version=get_version_from_git())},
        info if info else {},
    )
    filled = assoc_in(filled, [test_name, "_info"], info)

    if apply_formatter:
        return formatter(filled)
    else:
        return filled



from eth_typing import (
    Address,
)
from eth_utils.toolz import (
    curry,
)

from eth import (
    constants,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth.abc import (
    ComputationAPI,
)
from eth.vm import (
    mnemonics,
)
from eth.vm.forks.istanbul.storage import (
    GAS_SCHEDULE_EIP2200,
    sstore_eip2200_generic,
)
from eth.vm.logic.call import (
    CallByzantium,
    CallCodeEIP150,
    DelegateCallEIP150,
    StaticCall,
)
from eth.vm.logic.context import (
    consume_extcodecopy_word_cost,
    extcodecopy_execute,
    push_balance_of_address,
)
from eth.vm.logic.storage import (
    NetSStoreGasSchedule,
)
from eth.vm.logic.system import (
    Create2,
    CreateByzantium,
    CreateOpcodeStackData,
    selfdestruct_eip161_on_address,
)

from . import (
    constants as berlin_constants,
)


def _mark_address_warm(computation: ComputationAPI, address: Address) -> bool:
    """
    Mark the given address as warm if it was not previously.

    :return was_cold: True if the account was not previously accessed
        during this transaction
    """

    if computation.state.is_address_warm(address):
        return False
    else:
        computation.state.mark_address_warm(address)
        return True


def _account_load_cost(was_cold: bool) -> int:
    if was_cold:
        return berlin_constants.COLD_ACCOUNT_ACCESS_COST
    else:
        return berlin_constants.WARM_STORAGE_READ_COST


def _consume_gas_for_account_load(
    computation: ComputationAPI, address: Address, reason: str
) -> None:
    was_cold = _mark_address_warm(computation, address)
    gas_cost = _account_load_cost(was_cold)
    computation.consume_gas(gas_cost, reason=reason)


def _mark_storage_warm(computation: ComputationAPI, slot: int) -> bool:
    """
    :return was_cold: True if the storage slot was not previously accessed
        during this transaction
    """
    storage_address = computation.msg.storage_address
    if computation.state.is_storage_warm(storage_address, slot):
        return False
    else:
        computation.state.mark_storage_warm(storage_address, slot)
        return True


def balance_eip2929(computation: ComputationAPI) -> None:
    address = force_bytes_to_address(computation.stack_pop1_bytes())
    _consume_gas_for_account_load(computation, address, mnemonics.BALANCE)
    push_balance_of_address(address, computation)


def extcodesize_eip2929(computation: ComputationAPI) -> None:
    address = force_bytes_to_address(computation.stack_pop1_bytes())
    _consume_gas_for_account_load(computation, address, mnemonics.EXTCODEHASH)

    code_size = len(computation.state.get_code(address))
    computation.stack_push_int(code_size)


def extcodecopy_eip2929(computation: ComputationAPI) -> None:
    address, size = extcodecopy_execute(computation)
    consume_extcodecopy_word_cost(computation, size)
    _consume_gas_for_account_load(computation, address, mnemonics.EXTCODECOPY)


def extcodehash_eip2929(computation: ComputationAPI) -> None:
    """
    Return the code hash for a given address.
    EIP: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1052.md
    """
    address = force_bytes_to_address(computation.stack_pop1_bytes())
    state = computation.state

    _consume_gas_for_account_load(computation, address, mnemonics.EXTCODEHASH)

    if state.account_is_empty(address):
        computation.stack_push_bytes(constants.NULL_BYTE)
    else:
        computation.stack_push_bytes(state.get_code_hash(address))


def sload_eip2929(computation: ComputationAPI) -> None:
    slot = computation.stack_pop1_int()

    if _mark_storage_warm(computation, slot):
        gas_cost = berlin_constants.COLD_SLOAD_COST
    else:
        gas_cost = berlin_constants.WARM_STORAGE_READ_COST
    computation.consume_gas(gas_cost, reason=mnemonics.SLOAD)

    value = computation.state.get_storage(
        address=computation.msg.storage_address,
        slot=slot,
    )
    computation.stack_push_int(value)


GAS_SCHEDULE_EIP2929 = GAS_SCHEDULE_EIP2200._replace(
    sload_gas=berlin_constants.WARM_STORAGE_READ_COST,
    sstore_reset_gas=5000 - berlin_constants.COLD_SLOAD_COST,
)


@curry
def sstore_eip2929_generic(
    gas_schedule: NetSStoreGasSchedule,
    computation: ComputationAPI,
) -> int:
    slot = sstore_eip2200_generic(gas_schedule, computation)

    if _mark_storage_warm(computation, slot):
        gas_cost = berlin_constants.COLD_SLOAD_COST
        computation.consume_gas(
            gas_cost, reason=f"Implicit SLOAD during {mnemonics.SSTORE}"
        )

    return slot


sstore_eip2929 = sstore_eip2929_generic(GAS_SCHEDULE_EIP2929)


class LoadFeeByCacheWarmth:
    def get_account_load_fee(
        self,
        computation: ComputationAPI,
        code_address: Address,
    ) -> int:
        was_cold = _mark_address_warm(computation, code_address)
        return _account_load_cost(was_cold)


class CallEIP2929(LoadFeeByCacheWarmth, CallByzantium):
    pass


class CallCodeEIP2929(LoadFeeByCacheWarmth, CallCodeEIP150):
    pass


class DelegateCallEIP2929(LoadFeeByCacheWarmth, DelegateCallEIP150):
    pass


class StaticCallEIP2929(LoadFeeByCacheWarmth, StaticCall):
    pass


def selfdestruct_eip2929(computation: ComputationAPI) -> None:
    beneficiary = force_bytes_to_address(computation.stack_pop1_bytes())

    if _mark_address_warm(computation, beneficiary):
        gas_cost = berlin_constants.COLD_ACCOUNT_ACCESS_COST
        computation.consume_gas(
            gas_cost,
            reason=f"Implicit account load during {mnemonics.SELFDESTRUCT}",
        )

    selfdestruct_eip161_on_address(computation, beneficiary)


class CreateEIP2929(CreateByzantium):
    def generate_contract_address(
        self,
        stack_data: CreateOpcodeStackData,
        call_data: bytes,
        computation: ComputationAPI,
    ) -> Address:
        address = super().generate_contract_address(stack_data, call_data, computation)
        computation.state.mark_address_warm(address)
        return address


class Create2EIP2929(Create2):
    def generate_contract_address(
        self,
        stack_data: CreateOpcodeStackData,
        call_data: bytes,
        computation: ComputationAPI,
    ) -> Address:
        address = super().generate_contract_address(stack_data, call_data, computation)
        computation.state.mark_address_warm(address)
        return address



import copy
from typing import (
    Dict,
)

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.abc import (
    OpcodeAPI,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.frontier.opcodes import (
    FRONTIER_OPCODES,
)
from eth.vm.logic import (
    call,
)

NEW_OPCODES = {
    opcode_values.DELEGATECALL: call.DelegateCall.configure(
        __name__="opcode:DELEGATECALL",
        mnemonic=mnemonics.DELEGATECALL,
        gas_cost=constants.GAS_CALL,
    )(),
}


HOMESTEAD_OPCODES: Dict[int, OpcodeAPI] = merge(
    copy.deepcopy(FRONTIER_OPCODES), NEW_OPCODES
)



from typing import (
    Type,
)

from eth_utils import (
    encode_hex,
)

from eth.abc import (
    ComputationAPI,
    SignedTransactionAPI,
    TransactionExecutorAPI,
)
from eth.vm.forks.homestead.state import (
    HomesteadState,
    HomesteadTransactionExecutor,
)

from ._utils import (
    collect_touched_accounts,
)
from .computation import (
    SpuriousDragonComputation,
)


class SpuriousDragonTransactionExecutor(HomesteadTransactionExecutor):
    def finalize_computation(
        self, transaction: SignedTransactionAPI, computation: ComputationAPI
    ) -> ComputationAPI:
        computation = super().finalize_computation(transaction, computation)

        #
        # EIP161 state clearing
        #
        touched_accounts = collect_touched_accounts(computation)

        for account in touched_accounts:
            should_delete = self.vm_state.account_exists(
                account
            ) and self.vm_state.account_is_empty(account)
            if should_delete:
                self.vm_state.logger.debug2(
                    "CLEARING EMPTY ACCOUNT: %s",
                    encode_hex(account),
                )
                self.vm_state.delete_account(account)

        return computation


class SpuriousDragonState(HomesteadState):
    computation_class: Type[ComputationAPI] = SpuriousDragonComputation
    transaction_executor_class: Type[
        TransactionExecutorAPI
    ] = SpuriousDragonTransactionExecutor



from eth._utils.numeric import (
    ceil32,
)
from eth.vm.forks.berlin.logic import (
    Create2EIP2929,
    CreateEIP2929,
)
from eth.vm.logic.system import (
    CreateOpcodeStackData,
)

from .constants import (
    INITCODE_WORD_COST,
)


class CreateEIP3860(CreateEIP2929):
    def get_gas_cost(self, data: CreateOpcodeStackData) -> int:
        eip2929_gas_cost = super().get_gas_cost(data)
        eip3860_gas_cost = INITCODE_WORD_COST * ceil32(data.memory_length) // 32
        return eip2929_gas_cost + eip3860_gas_cost


class Create2EIP3860(Create2EIP2929):
    def get_gas_cost(self, data: CreateOpcodeStackData) -> int:
        eip2929_gas_cost = super().get_gas_cost(data)
        eip3860_gas_cost = INITCODE_WORD_COST * ceil32(data.memory_length) // 32
        return eip2929_gas_cost + eip3860_gas_cost



from eth_utils import (
    to_tuple,
)

from eth.typing import (
    AccountDiff,
    AccountState,
)
from eth.vm.state import (
    BaseState,
)


@to_tuple
def diff_state(expected_state: AccountState, state: BaseState) -> AccountDiff:
    for account, account_data in sorted(expected_state.items()):
        expected_balance = account_data["balance"]
        expected_nonce = account_data["nonce"]
        expected_code = account_data["code"]

        actual_nonce = state.get_nonce(account)
        actual_code = state.get_code(account)
        actual_balance = state.get_balance(account)

        if actual_nonce != expected_nonce:
            yield (account, "nonce", actual_nonce, expected_nonce)
        if actual_code != expected_code:
            yield (account, "code", actual_code, expected_code)
        if actual_balance != expected_balance:
            yield (account, "balance", actual_balance, expected_balance)

        for slot, expected_storage_value in sorted(account_data["storage"].items()):
            actual_storage_value = state.get_storage(account, slot)
            if actual_storage_value != expected_storage_value:
                yield (
                    account,
                    f"storage[{slot}]",
                    actual_storage_value,
                    expected_storage_value,
                )
